@@ -12,16 +12,16 @@
 
 #ifdef UA_ENABLE_ENCRYPTION_MBEDTLS
 
+#include "securitypolicy_mbedtls_common.h"
+
 #include <mbedtls/x509.h>
 #include <mbedtls/oid.h>
 #include <mbedtls/x509_crt.h>
+#include <mbedtls/entropy.h>
 #include <mbedtls/error.h>
 #include <mbedtls/version.h>
 #include <string.h>
 
-#define UA_MBEDTLS_MAX_CHAIN_LENGTH 10
-#define UA_MBEDTLS_MAX_DN_LENGTH 256
-
 /* Find binary substring. Taken and adjusted from
  * http://tungchingkai.blogspot.com/2011/07/binary-strstr.html */
 
@@ -42,7 +42,7 @@ UA_Bstrstr(const unsigned char *s1, size_t l1, const unsigned char *s2, size_t l
     const unsigned char *ss2 = s2;
     /* handle special case */
     if(l1 == 0)
-        return NULL;
+        return (NULL);
     if(l2 == 0)
         return s1;
 
@@ -88,6 +88,7 @@ typedef struct {
     UA_String trustListFolder;
     UA_String issuerListFolder;
     UA_String revocationListFolder;
+    UA_String rejectedListFolder;
 
     mbedtls_x509_crt certificateTrustList;
     mbedtls_x509_crt certificateIssuerList;
@@ -107,7 +108,7 @@ fileNamesFromFolder(const UA_String *folder, size_t *pathsSize, UA_String **path
 
     memcpy(buf, folder->data, folder->length);
     buf[folder->length] = 0;
-    
+
     DIR *dir = opendir(buf);
     if(!dir)
         return UA_STATUSCODE_BADINTERNALERROR;
@@ -146,14 +147,14 @@ fileNamesFromFolder(const UA_String *folder, size_t *pathsSize, UA_String **path
 }
 
 static UA_StatusCode
-reloadCertificates(CertInfo *ci) {
+reloadCertificates(const UA_CertificateVerification *cv, CertInfo *ci) {
     UA_StatusCode retval = UA_STATUSCODE_GOOD;
     int err = 0;
     int internalErrorFlag = 0;
 
     /* Load the trustlists */
     if(ci->trustListFolder.length > 0) {
-        UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "Reloading the trust-list");
+        UA_LOG_INFO(cv->logging, UA_LOGCATEGORY_SERVER, "Reloading the trust-list");
         mbedtls_x509_crt_free(&ci->certificateTrustList);
         mbedtls_x509_crt_init(&ci->certificateTrustList);
 
@@ -162,12 +163,12 @@ reloadCertificates(CertInfo *ci) {
         f[ci->trustListFolder.length] = 0;
         err = mbedtls_x509_crt_parse_path(&ci->certificateTrustList, f);
         if(err == 0) {
-            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
+            UA_LOG_INFO(cv->logging, UA_LOGCATEGORY_SERVER,
                         "Loaded certificate from %s", f);
         } else {
             char errBuff[300];
             mbedtls_strerror(err, errBuff, 300);
-            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
+            UA_LOG_INFO(cv->logging, UA_LOGCATEGORY_SERVER,
                         "Failed to load certificate from %s, mbedTLS error: %s (error code: %d)", f, errBuff, err);
             internalErrorFlag = 1;
         }
@@ -175,7 +176,7 @@ reloadCertificates(CertInfo *ci) {
 
     /* Load the revocationlists */
     if(ci->revocationListFolder.length > 0) {
-        UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "Reloading the revocation-list");
+        UA_LOG_INFO(cv->logging, UA_LOGCATEGORY_SERVER, "Reloading the revocation-list");
         size_t pathsSize = 0;
         UA_String *paths = NULL;
         retval = fileNamesFromFolder(&ci->revocationListFolder, &pathsSize, &paths);
@@ -189,13 +190,13 @@ reloadCertificates(CertInfo *ci) {
             f[paths[i].length] = 0;
             err = mbedtls_x509_crl_parse_file(&ci->certificateRevocationList, f);
             if(err == 0) {
-                UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
+                UA_LOG_INFO(cv->logging, UA_LOGCATEGORY_SERVER,
                             "Loaded certificate from %.*s",
                             (int)paths[i].length, paths[i].data);
             } else {
                 char errBuff[300];
                 mbedtls_strerror(err, errBuff, 300);
-                UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
+                UA_LOG_INFO(cv->logging, UA_LOGCATEGORY_SERVER,
                             "Failed to load certificate from %.*s, mbedTLS error: %s (error code: %d)",
                             (int)paths[i].length, paths[i].data, errBuff, err);
                 internalErrorFlag = 1;
@@ -208,7 +209,7 @@ reloadCertificates(CertInfo *ci) {
 
     /* Load the issuerlists */
     if(ci->issuerListFolder.length > 0) {
-        UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "Reloading the issuer-list");
+        UA_LOG_INFO(cv->logging, UA_LOGCATEGORY_SERVER, "Reloading the issuer-list");
         mbedtls_x509_crt_free(&ci->certificateIssuerList);
         mbedtls_x509_crt_init(&ci->certificateIssuerList);
         char f[PATH_MAX];
@@ -216,23 +217,29 @@ reloadCertificates(CertInfo *ci) {
         f[ci->issuerListFolder.length] = 0;
         err = mbedtls_x509_crt_parse_path(&ci->certificateIssuerList, f);
         if(err == 0) {
-            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
+            UA_LOG_INFO(cv->logging, UA_LOGCATEGORY_SERVER,
                         "Loaded certificate from %s", f);
         } else {
             char errBuff[300];
             mbedtls_strerror(err, errBuff, 300);
-            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
-                        "Failed to load certificate from %s, mbedTLS error: %s (error code: %d)", 
+            UA_LOG_INFO(cv->logging, UA_LOGCATEGORY_SERVER,
+                        "Failed to load certificate from %s, mbedTLS error: %s (error code: %d)",
                         f, errBuff, err);
             internalErrorFlag = 1;
         }
     }
 
-    return (internalErrorFlag) ? UA_STATUSCODE_BADINTERNALERROR : retval;
+    if(internalErrorFlag) {
+        retval = UA_STATUSCODE_BADINTERNALERROR;
+    }
+    return retval;
 }
 
 #endif
 
+#define UA_MBEDTLS_MAX_CHAIN_LENGTH 10
+#define UA_MBEDTLS_MAX_DN_LENGTH 256
+
 /* We need to access some private fields below */
 #ifndef MBEDTLS_PRIVATE
 #define MBEDTLS_PRIVATE(x) x
@@ -265,13 +272,6 @@ mbedtlsSameName(UA_String name, const mbedtls_x509_name *name2) {
     return UA_String_equal(&name, &nameString);
 }
 
-static UA_Boolean
-mbedtlsSameBuf(mbedtls_x509_buf *a, mbedtls_x509_buf *b) {
-    if(a->len != b->len)
-        return false;
-    return (memcmp(a->p, b->p, a->len) == 0);
-}
-
 /* Return the first matching issuer candidate AFTER prev.
  * This can return the cert itself if self-signed. */
 static mbedtls_x509_crt *
@@ -295,47 +295,27 @@ mbedtlsFindNextIssuer(CertInfo *ci, mbedtls_x509_crt *stack,
                mbedtls_pk_can_do(&i->pk, cert->MBEDTLS_PRIVATE(sig_pk)))
                 return i;
         }
-
-        /* Switch from the stack that came with the cert to the issuer list and
-         * then to the trust list. */
-        if(stack == &ci->certificateTrustList)
-            stack = NULL;
-        else if(stack == &ci->certificateIssuerList)
-            stack = &ci->certificateTrustList;
-        else
-            stack = &ci->certificateIssuerList;
+        /* Switch from the stack that came with the cert to the ctx->skIssue list */
+        stack = (stack != &ci->certificateIssuerList) ? &ci->certificateIssuerList : NULL;
     } while(stack);
     return NULL;
 }
 
-static UA_StatusCode
+static UA_Boolean
 mbedtlsCheckRevoked(CertInfo *ci, mbedtls_x509_crt *cert) {
-    /* Parse the Issuer Name */
     char inbuf[UA_MBEDTLS_MAX_DN_LENGTH];
     int nameLen = mbedtls_x509_dn_gets(inbuf, UA_MBEDTLS_MAX_DN_LENGTH, &cert->issuer);
     if(nameLen < 0)
-        return UA_STATUSCODE_BADINTERNALERROR;
+        return true;
     UA_String issuerName = {(size_t)nameLen, (UA_Byte*)inbuf};
-
-    if(ci->certificateRevocationList.raw.len == 0) {
-        UA_LOG_WARNING(UA_Log_Stdout, UA_LOGCATEGORY_SECURITYPOLICY,
-                       "Zero revocation lists have been loaded. "
-                       "This seems intentional - omitting the check.");
-        return UA_STATUSCODE_GOOD;
-    }
-
-    /* Loop over the crl and match the Issuer Name */
-    UA_StatusCode res = UA_STATUSCODE_BADCERTIFICATEREVOCATIONUNKNOWN;
     for(mbedtls_x509_crl *crl = &ci->certificateRevocationList; crl; crl = crl->next) {
         /* Is the CRL for certificates from the cert issuer?
          * Is the serial number of the certificate contained in the CRL? */
-        if(mbedtlsSameName(issuerName, &crl->issuer)) {
-            if(mbedtls_x509_crt_is_revoked(cert, crl) != 0)
-                return UA_STATUSCODE_BADCERTIFICATEREVOKED;
-            res = UA_STATUSCODE_GOOD; /* There was at least one crl that did not revoke (so far) */
-        }
+        if(mbedtlsSameName(issuerName, &crl->issuer) &&
+           mbedtls_x509_crt_is_revoked(cert, crl) != 0)
+            return true;
     }
-    return res;
+    return false;
 }
 
 /* Verify that the public key of the issuer was used to sign the certificate */
@@ -374,6 +354,11 @@ mbedtlsVerifyChain(CertInfo *ci, mbedtls_x509_crt *stack, mbedtls_x509_crt **old
         return (depth == 0) ? UA_STATUSCODE_BADCERTIFICATETIMEINVALID :
             UA_STATUSCODE_BADCERTIFICATEISSUERTIMEINVALID;
 
+    /* Verification Step: Revocation Check */
+    if(mbedtlsCheckRevoked(ci, cert))
+        return (depth == 0) ? UA_STATUSCODE_BADCERTIFICATEREVOKED :
+            UA_STATUSCODE_BADCERTIFICATEISSUERREVOKED;
+
     /* Return the most specific error code. BADCERTIFICATECHAININCOMPLETE is
      * returned only if all possible chains are incomplete. */
     mbedtls_x509_crt *issuer = NULL;
@@ -403,28 +388,16 @@ mbedtlsVerifyChain(CertInfo *ci, mbedtls_x509_crt *stack, mbedtls_x509_crt **old
          * chain. We check whether the certificate is trusted below. This is the
          * only place where we return UA_STATUSCODE_BADCERTIFICATEUNTRUSTED.
          * This signals that the chain is complete (but can be still
-         * untrusted).
-         *
-         * Break here as we have reached the end of the chain. Omit the
-         * Revocation Check for self-signed certificates. */
-        if(issuer == cert || mbedtlsSameBuf(&cert->tbs, &issuer->tbs)) {
+         * untrusted). */
+        if(issuer == cert || (cert->tbs.len == issuer->tbs.len &&
+                              memcmp(cert->tbs.p, issuer->tbs.p, cert->tbs.len) == 0)) {
             ret = UA_STATUSCODE_BADCERTIFICATEUNTRUSTED;
-            break;
-        }
-
-        /* Verification Step: Revocation Check */
-        ret = mbedtlsCheckRevoked(ci, cert);
-        if(depth > 0) {
-            if(ret == UA_STATUSCODE_BADCERTIFICATEREVOKED)
-                ret = UA_STATUSCODE_BADCERTIFICATEISSUERREVOKED;
-            if(ret == UA_STATUSCODE_BADCERTIFICATEREVOCATIONUNKNOWN)
-                ret = UA_STATUSCODE_BADCERTIFICATEISSUERREVOCATIONUNKNOWN;
-        }
-        if(ret != UA_STATUSCODE_GOOD)
             continue;
+        }
 
-        /* Detect (endless) loops of issuers */
-        for(int i = 0; i < depth; i++) {
+        /* Detect (endless) loops of issuers. The last one can be skipped by the
+         * check for self-signed just before. */
+        for(int i = 0; i < depth - 1; i++) {
             if(old_issuers[i] == issuer)
                 return UA_STATUSCODE_BADCERTIFICATECHAININCOMPLETE;
         }
@@ -439,7 +412,8 @@ mbedtlsVerifyChain(CertInfo *ci, mbedtls_x509_crt *stack, mbedtls_x509_crt **old
      * certificate "on the way down". Can we trust this certificate? */
     if(ret == UA_STATUSCODE_BADCERTIFICATEUNTRUSTED) {
         for(mbedtls_x509_crt *t = &ci->certificateTrustList; t; t = t->next) {
-            if(mbedtlsSameBuf(&cert->tbs, &t->tbs))
+            if(cert->tbs.len == t->tbs.len &&
+               memcmp(cert->tbs.p, t->tbs.p, cert->tbs.len) == 0)
                 return UA_STATUSCODE_GOOD;
         }
     }
@@ -450,16 +424,16 @@ mbedtlsVerifyChain(CertInfo *ci, mbedtls_x509_crt *stack, mbedtls_x509_crt **old
 /* This follows Part 6, 6.1.3 Determining if a Certificate is trusted.
  * It defines a sequence of steps for certificate verification. */
 static UA_StatusCode
-certificateVerification_verify(void *verificationContext,
-                                const UA_ByteString *certificate) {
-    if(!verificationContext || !certificate)
+certificateVerification_verify(const UA_CertificateVerification *cv,
+                               const UA_ByteString *certificate) {
+    if(!cv || !certificate)
         return UA_STATUSCODE_BADINTERNALERROR;
 
     UA_StatusCode ret = UA_STATUSCODE_GOOD;
-    CertInfo *ci = (CertInfo*)verificationContext;
+    CertInfo *ci = (CertInfo*)cv->context;
 
 #ifdef __linux__ /* Reload certificates if folder paths are specified */
-    ret = reloadCertificates(ci);
+    ret = reloadCertificates(cv, ci);
     if(ret != UA_STATUSCODE_GOOD)
         return ret;
 #endif
@@ -497,10 +471,13 @@ certificateVerification_verify(void *verificationContext,
 }
 
 static UA_StatusCode
-certificateVerification_verifyApplicationURI(void *verificationContext,
+certificateVerification_verifyApplicationURI(const UA_CertificateVerification *cv,
                                              const UA_ByteString *certificate,
                                              const UA_String *applicationURI) {
-    CertInfo *ci = (CertInfo*)verificationContext;
+    CertInfo *ci;
+    if(!cv)
+        return UA_STATUSCODE_BADINTERNALERROR;
+    ci = (CertInfo*)cv->context;
     if(!ci)
         return UA_STATUSCODE_BADINTERNALERROR;
 
@@ -537,10 +514,53 @@ certificateVerification_clear(UA_CertificateVerification *cv) {
     UA_String_clear(&ci->trustListFolder);
     UA_String_clear(&ci->issuerListFolder);
     UA_String_clear(&ci->revocationListFolder);
+#ifdef UA_ENABLE_CERT_REJECTED_DIR
+    UA_String_clear(&ci->rejectedListFolder);
+#endif
     UA_free(ci);
     cv->context = NULL;
 }
 
+static UA_StatusCode
+getCertificate_ExpirationDate(UA_DateTime *expiryDateTime, 
+                              UA_ByteString *certificate) {
+    mbedtls_x509_crt publicKey;
+    mbedtls_x509_crt_init(&publicKey);
+    int mbedErr = mbedtls_x509_crt_parse(&publicKey, certificate->data, certificate->length);
+    if(mbedErr)
+        return UA_STATUSCODE_BADINTERNALERROR;
+    UA_DateTimeStruct ts;
+    ts.year = (UA_Int16)publicKey.valid_to.year;
+    ts.month = (UA_UInt16)publicKey.valid_to.mon;
+    ts.day = (UA_UInt16)publicKey.valid_to.day;
+    ts.hour = (UA_UInt16)publicKey.valid_to.hour;
+    ts.min = (UA_UInt16)publicKey.valid_to.min;
+    ts.sec = (UA_UInt16)publicKey.valid_to.sec;
+    ts.milliSec = 0;
+    ts.microSec = 0;
+    ts.nanoSec = 0;
+    *expiryDateTime = UA_DateTime_fromStruct(ts);
+    mbedtls_x509_crt_free(&publicKey);
+    return UA_STATUSCODE_GOOD;
+}
+
+static UA_StatusCode
+getCertificate_SubjectName(UA_String *subjectName,
+                           UA_ByteString *certificate) {
+    mbedtls_x509_crt publicKey;
+    mbedtls_x509_crt_init(&publicKey);
+    int mbedErr = mbedtls_x509_crt_parse(&publicKey, certificate->data, certificate->length);
+    if(mbedErr)
+        return UA_STATUSCODE_BADINTERNALERROR;
+    char buf[1024];
+    int res = mbedtls_x509_dn_gets(buf, 1024, &publicKey.subject);
+    mbedtls_x509_crt_free(&publicKey);
+    if(res < 0)
+        return UA_STATUSCODE_BADINTERNALERROR;
+    UA_String tmp = {(size_t)res, (UA_Byte*)buf};
+    return UA_String_copy(&tmp, subjectName);
+}
+
 UA_StatusCode
 UA_CertificateVerification_Trustlist(UA_CertificateVerification *cv,
                                      const UA_ByteString *certificateTrustList,
@@ -549,6 +569,15 @@ UA_CertificateVerification_Trustlist(UA_CertificateVerification *cv,
                                      size_t certificateIssuerListSize,
                                      const UA_ByteString *certificateRevocationList,
                                      size_t certificateRevocationListSize) {
+
+    if(cv == NULL) {
+        return UA_STATUSCODE_BADINTERNALERROR;
+    }
+
+    /* Clear if the plugin is already initialized */
+    if(cv->clear)
+        cv->clear(cv);
+
     CertInfo *ci = (CertInfo*)UA_malloc(sizeof(CertInfo));
     if(!ci)
         return UA_STATUSCODE_BADOUTOFMEMORY;
@@ -561,6 +590,8 @@ UA_CertificateVerification_Trustlist(UA_CertificateVerification *cv,
     cv->verifyCertificate = certificateVerification_verify;
     cv->clear = certificateVerification_clear;
     cv->verifyApplicationURI = certificateVerification_verifyApplicationURI;
+    cv->getExpirationDate = getCertificate_ExpirationDate;
+    cv->getSubjectName = getCertificate_SubjectName;
 
     int err;
     UA_ByteString data;
@@ -602,11 +633,32 @@ error:
 
 #ifdef __linux__ /* Linux only so far */
 
+#ifdef UA_ENABLE_CERT_REJECTED_DIR
+UA_StatusCode
+UA_CertificateVerification_CertFolders(UA_CertificateVerification *cv,
+                                       const char *trustListFolder,
+                                       const char *issuerListFolder,
+                                       const char *revocationListFolder,
+                                       const char *rejectedListFolder) {
+#else
 UA_StatusCode
 UA_CertificateVerification_CertFolders(UA_CertificateVerification *cv,
                                        const char *trustListFolder,
                                        const char *issuerListFolder,
                                        const char *revocationListFolder) {
+#endif
+    UA_StatusCode ret;
+    if(cv == NULL) {
+        return UA_STATUSCODE_BADINTERNALERROR;
+    }
+    if(cv->logging == NULL) {
+        return UA_STATUSCODE_BADINTERNALERROR;
+    }
+
+    /* Clear if the plugin is already initialized */
+    if(cv->clear)
+        cv->clear(cv);
+
     CertInfo *ci = (CertInfo*)UA_malloc(sizeof(CertInfo));
     if(!ci)
         return UA_STATUSCODE_BADOUTOFMEMORY;
@@ -620,17 +672,80 @@ UA_CertificateVerification_CertFolders(UA_CertificateVerification *cv,
     ci->trustListFolder = UA_STRING_ALLOC(trustListFolder);
     ci->issuerListFolder = UA_STRING_ALLOC(issuerListFolder);
     ci->revocationListFolder = UA_STRING_ALLOC(revocationListFolder);
-
-    reloadCertificates(ci);
+#ifdef UA_ENABLE_CERT_REJECTED_DIR
+    ci->rejectedListFolder = UA_STRING_ALLOC(rejectedListFolder);
+#endif
 
     cv->context = (void*)ci;
     cv->verifyCertificate = certificateVerification_verify;
     cv->clear = certificateVerification_clear;
     cv->verifyApplicationURI = certificateVerification_verifyApplicationURI;
 
-    return UA_STATUSCODE_GOOD;
+    ret = reloadCertificates(cv, ci);
+
+    return ret;
 }
 
 #endif
 
+UA_StatusCode
+UA_PKI_decryptPrivateKey(const UA_ByteString privateKey,
+                         const UA_ByteString password,
+                         UA_ByteString *outDerKey) {
+    if(!outDerKey)
+        return UA_STATUSCODE_BADINTERNALERROR;
+
+    if (privateKey.length == 0) {
+        *outDerKey = UA_BYTESTRING_NULL;
+        return UA_STATUSCODE_BADINVALIDARGUMENT;
+    }
+
+    /* Already in DER format -> return verbatim */
+    if(privateKey.length > 1 && privateKey.data[0] == 0x30 && privateKey.data[1] == 0x82)
+        return UA_ByteString_copy(&privateKey, outDerKey);
+
+    /* Create a null-terminated string */
+    UA_ByteString nullTerminatedKey = UA_mbedTLS_CopyDataFormatAware(&privateKey);
+    if(nullTerminatedKey.length != privateKey.length + 1)
+        return UA_STATUSCODE_BADOUTOFMEMORY;
+
+    /* Create the private-key context */
+    mbedtls_pk_context ctx;
+    mbedtls_pk_init(&ctx);
+#if MBEDTLS_VERSION_NUMBER >= 0x02060000 && MBEDTLS_VERSION_NUMBER < 0x03000000
+    int err = mbedtls_pk_parse_key(&ctx, nullTerminatedKey.data,
+                                   nullTerminatedKey.length,
+                                   password.data, password.length);
+#else
+    mbedtls_entropy_context entropy;
+    mbedtls_entropy_init(&entropy);
+    int err = mbedtls_pk_parse_key(&ctx, nullTerminatedKey.data,
+                                   nullTerminatedKey.length,
+                                   password.data, password.length,
+                                   mbedtls_entropy_func, &entropy);
+    mbedtls_entropy_free(&entropy);
+#endif
+    UA_ByteString_clear(&nullTerminatedKey);
+    if(err != 0) {
+        mbedtls_pk_free(&ctx);
+        return UA_STATUSCODE_BADSECURITYCHECKSFAILED;
+    }
+
+    /* Write the DER-encoded key into a local buffer */
+    unsigned char buf[1 << 14];
+    size_t pos = (size_t)mbedtls_pk_write_key_der(&ctx, buf, sizeof(buf));
+
+    /* Allocate memory */
+    UA_StatusCode res = UA_ByteString_allocBuffer(outDerKey, pos);
+    if(res != UA_STATUSCODE_GOOD) {
+        mbedtls_pk_free(&ctx);
+        return res;
+    }
+
+    /* Copy to the output */
+    memcpy(outDerKey->data, &buf[sizeof(buf) - pos], pos);
+    mbedtls_pk_free(&ctx);
+    return UA_STATUSCODE_GOOD;
+}
+
 #endif /* UA_ENABLE_ENCRYPTION_MBEDTLS */
