@@ -7,7 +7,7 @@
  *    Copyright 2018 (c) HMS Industrial Networks AB (Author: Jonas Green)
  *    Copyright 2020 (c) Wind River Systems, Inc.
  *    Copyright 2020 (c) basysKom GmbH
- * 
+ *
  */
 
 #include <open62541/plugin/securitypolicy_default.h>
@@ -98,21 +98,17 @@ asym_getLocalSignatureSize_sp_basic128rsa15(const Basic128Rsa15_ChannelContext *
 #else
     return mbedtls_rsa_get_len(mbedtls_pk_rsa(cc->policyContext->localPrivateKey));
 #endif
-
-
 }
 
 static size_t
 asym_getRemoteSignatureSize_sp_basic128rsa15(const Basic128Rsa15_ChannelContext *cc) {
     if(cc == NULL)
         return 0;
-
 #if MBEDTLS_VERSION_NUMBER >= 0x02060000 && MBEDTLS_VERSION_NUMBER < 0x03000000
     return mbedtls_pk_rsa(cc->remoteCertificate.pk)->len;
 #else
     return mbedtls_rsa_get_len(mbedtls_pk_rsa(cc->remoteCertificate.pk));
 #endif
-
 }
 
 static UA_StatusCode
@@ -125,23 +121,18 @@ asym_encrypt_sp_basic128rsa15(Basic128Rsa15_ChannelContext *cc,
     mbedtls_rsa_set_padding(remoteRsaContext, MBEDTLS_RSA_PKCS_V15, MBEDTLS_MD_NONE);
 
 #if MBEDTLS_VERSION_NUMBER >= 0x02060000 && MBEDTLS_VERSION_NUMBER < 0x03000000
-    size_t plainTextBlockSize = remoteRsaContext->len - UA_SECURITYPOLICY_BASIC128RSA15_RSAPADDING_LEN;
+    size_t keylen = remoteRsaContext->len;
 #else
     size_t keylen = mbedtls_rsa_get_len(remoteRsaContext);
-    size_t plainTextBlockSize = mbedtls_rsa_get_len(remoteRsaContext) -
-        UA_SECURITYPOLICY_BASIC128RSA15_RSAPADDING_LEN;
 #endif
-
+    size_t plainTextBlockSize =
+        keylen - UA_SECURITYPOLICY_BASIC128RSA15_RSAPADDING_LEN;
     if(data->length % plainTextBlockSize != 0)
         return UA_STATUSCODE_BADINTERNALERROR;
 
     size_t blocks = data->length / plainTextBlockSize;
     UA_ByteString encrypted;
-#if MBEDTLS_VERSION_NUMBER >= 0x02060000 && MBEDTLS_VERSION_NUMBER < 0x03000000
-    UA_StatusCode retval = UA_ByteString_allocBuffer(&encrypted, blocks * remoteRsaContext->len);
-#else
     UA_StatusCode retval = UA_ByteString_allocBuffer(&encrypted, blocks * keylen);
-#endif
     if(retval != UA_STATUSCODE_GOOD)
         return retval;
 
@@ -215,12 +206,12 @@ asym_decrypt_sp_basic128rsa15(Basic128Rsa15_ChannelContext *cc,
 
 static size_t
 asym_getLocalEncryptionKeyLength_sp_basic128rsa15(const Basic128Rsa15_ChannelContext *cc) {
+    if(cc == NULL)
+        return 0;
 #if MBEDTLS_VERSION_NUMBER >= 0x02060000 && MBEDTLS_VERSION_NUMBER < 0x03000000
     mbedtls_rsa_context *const rsaContext = mbedtls_pk_rsa(cc->remoteCertificate.pk);
     return rsaContext->len;
 #else
-    if(cc == NULL)
-        return 0;
     return mbedtls_rsa_get_len(mbedtls_pk_rsa(cc->remoteCertificate.pk));
 #endif
 }
@@ -236,20 +227,18 @@ asym_getRemoteBlockSize_sp_basic128rsa15(const Basic128Rsa15_ChannelContext *cc)
     mbedtls_rsa_context *const rsaContext = mbedtls_pk_rsa(cc->remoteCertificate.pk);
     return rsaContext->len;
 #else
-    if(cc == NULL)
-        return 0;
     return mbedtls_rsa_get_len(mbedtls_pk_rsa(cc->remoteCertificate.pk));
 #endif
 }
 
 static size_t
 asym_getRemotePlainTextBlockSize_sp_basic128rsa15(const Basic128Rsa15_ChannelContext *cc) {
+    if(cc == NULL)
+        return 0;
 #if MBEDTLS_VERSION_NUMBER >= 0x02060000 && MBEDTLS_VERSION_NUMBER < 0x03000000
     mbedtls_rsa_context *const rsaContext = mbedtls_pk_rsa(cc->remoteCertificate.pk);
     return rsaContext->len - UA_SECURITYPOLICY_BASIC128RSA15_RSAPADDING_LEN;
 #else
-    if(cc == NULL)
-        return 0;
     return mbedtls_rsa_get_len(mbedtls_pk_rsa(cc->remoteCertificate.pk)) -
         UA_SECURITYPOLICY_BASIC128RSA15_RSAPADDING_LEN;
 #endif
