@@ -29,28 +29,28 @@ typedef struct {
 } Policy_Context_Basic256;
 
 typedef struct {
-    UA_ByteString             localSymSigningKey;  
-    UA_ByteString             localSymEncryptingKey; 
-    UA_ByteString             localSymIv; 
+    UA_ByteString             localSymSigningKey;
+    UA_ByteString             localSymEncryptingKey;
+    UA_ByteString             localSymIv;
     UA_ByteString             remoteSymSigningKey;
     UA_ByteString             remoteSymEncryptingKey;
     UA_ByteString             remoteSymIv;
 
     Policy_Context_Basic256 * policyContext;
     UA_ByteString             remoteCertificate;
-    X509 *                    remoteCertificateX509;   
+    X509 *                    remoteCertificateX509;
 } Channel_Context_Basic256;
 
-static UA_StatusCode 
+static UA_StatusCode
 UA_Policy_Basic256_New_Context (UA_SecurityPolicy * securityPolicy,
                                 const UA_ByteString localPrivateKey,
                                 const UA_Logger *   logger) {
-    Policy_Context_Basic256 * context = (Policy_Context_Basic256 *) 
+    Policy_Context_Basic256 * context = (Policy_Context_Basic256 *)
                                     UA_malloc (sizeof (Policy_Context_Basic256));
     if (context == NULL) {
         return UA_STATUSCODE_BADOUTOFMEMORY;
     }
-    
+
     context->localPrivateKey = UA_OpenSSL_LoadPrivateKey(&localPrivateKey);
 
     if (!context->localPrivateKey) {
@@ -65,14 +65,14 @@ UA_Policy_Basic256_New_Context (UA_SecurityPolicy * securityPolicy,
     if (retval != UA_STATUSCODE_GOOD) {
         EVP_PKEY_free(context->localPrivateKey);
         UA_free (context);
-        return retval; 
+        return retval;
     }
 
     context->logger = logger;
     securityPolicy->policyContext = context;
 
     return UA_STATUSCODE_GOOD;
-}                                    
+}
 
 static void
 UA_Policy_Basic256_Clear_Context (UA_SecurityPolicy *policy) {
@@ -81,29 +81,78 @@ UA_Policy_Basic256_Clear_Context (UA_SecurityPolicy *policy) {
     }
     UA_ByteString_clear(&policy->localCertificate);
 
-    Policy_Context_Basic256 * ctx = (Policy_Context_Basic256 *) policy->policyContext;    
+    Policy_Context_Basic256 * ctx = (Policy_Context_Basic256 *) policy->policyContext;
     if (ctx == NULL) {
-        return; 
+        return;
     }
 
     /* delete all allocated members in the context */
 
     EVP_PKEY_free(ctx->localPrivateKey);
     UA_ByteString_clear(&ctx->localCertThumbprint);
-    UA_free (ctx);   
+    UA_free (ctx);
 
-    return;         
+    return;
+}
+
+static UA_StatusCode
+updateCertificateAndPrivateKey_sp_basic256(UA_SecurityPolicy *securityPolicy,
+                                           const UA_ByteString newCertificate,
+                                           const UA_ByteString newPrivateKey) {
+    if(securityPolicy == NULL)
+        return UA_STATUSCODE_BADINTERNALERROR;
+
+    if(securityPolicy->policyContext == NULL)
+        return UA_STATUSCODE_BADINTERNALERROR;
+
+    Policy_Context_Basic256 *pc =
+        (Policy_Context_Basic256 *)securityPolicy->policyContext;
+
+    UA_ByteString_clear(&securityPolicy->localCertificate);
+
+    UA_StatusCode retval = UA_OpenSSL_LoadLocalCertificate(
+        &newCertificate, &securityPolicy->localCertificate);
+
+    if(retval != UA_STATUSCODE_GOOD)
+        return retval;
+
+    /* Set the new private key */
+    EVP_PKEY_free(pc->localPrivateKey);
+
+    pc->localPrivateKey = UA_OpenSSL_LoadPrivateKey(&newPrivateKey);
+
+    if(!pc->localPrivateKey) {
+        retval = UA_STATUSCODE_BADSECURITYCHECKSFAILED;
+        goto error;
+    }
+
+    UA_ByteString_clear(&pc->localCertThumbprint);
+
+    retval = UA_Openssl_X509_GetCertificateThumbprint(&securityPolicy->localCertificate,
+                                                      &pc->localCertThumbprint, true);
+    if(retval != UA_STATUSCODE_GOOD) {
+        goto error;
+    }
+
+    return retval;
+
+error:
+    UA_LOG_ERROR(securityPolicy->logger, UA_LOGCATEGORY_SECURITYPOLICY,
+                 "Could not update certificate and private key");
+    if(securityPolicy->policyContext != NULL)
+        UA_Policy_Basic256_Clear_Context(securityPolicy);
+    return retval;
 }
 
 /* create the channel context */
 
-static UA_StatusCode 
+static UA_StatusCode
 UA_ChannelModule_Basic256_New_Context (const UA_SecurityPolicy * securityPolicy,
                                        const UA_ByteString *     remoteCertificate,
                                        void **                   channelContext) {
-    if (securityPolicy == NULL || remoteCertificate == NULL || 
+    if (securityPolicy == NULL || remoteCertificate == NULL ||
         channelContext == NULL) {
-        return UA_STATUSCODE_BADINTERNALERROR;                                  
+        return UA_STATUSCODE_BADINTERNALERROR;
         }
     Channel_Context_Basic256 * context = (Channel_Context_Basic256 *)
             UA_malloc (sizeof (Channel_Context_Basic256));
@@ -118,7 +167,7 @@ UA_ChannelModule_Basic256_New_Context (const UA_SecurityPolicy * securityPolicy,
     UA_ByteString_init(&context->remoteSymEncryptingKey);
     UA_ByteString_init(&context->remoteSymIv);
 
-    UA_StatusCode retval = UA_copyCertificate (&context->remoteCertificate, 
+    UA_StatusCode retval = UA_copyCertificate (&context->remoteCertificate,
                                                remoteCertificate);
     if (retval != UA_STATUSCODE_GOOD) {
         UA_free (context);
@@ -128,18 +177,18 @@ UA_ChannelModule_Basic256_New_Context (const UA_SecurityPolicy * securityPolicy,
     /* decode to X509 */
     context->remoteCertificateX509 = UA_OpenSSL_LoadCertificate(&context->remoteCertificate);
     if (context->remoteCertificateX509 == NULL) {
-        UA_ByteString_clear (&context->remoteCertificate); 
+        UA_ByteString_clear (&context->remoteCertificate);
         UA_free (context);
         return UA_STATUSCODE_BADCERTIFICATECHAININCOMPLETE;
     }
 
-    context->policyContext = (Policy_Context_Basic256 *) 
+    context->policyContext = (Policy_Context_Basic256 *)
                              (securityPolicy->policyContext);
 
     *channelContext = context;
 
-    UA_LOG_INFO (securityPolicy->logger, 
-                 UA_LOGCATEGORY_SECURITYPOLICY, 
+    UA_LOG_INFO (securityPolicy->logger,
+                 UA_LOGCATEGORY_SECURITYPOLICY,
                  "The basic256 security policy channel with openssl is created.");
 
     return UA_STATUSCODE_GOOD;
@@ -147,29 +196,29 @@ UA_ChannelModule_Basic256_New_Context (const UA_SecurityPolicy * securityPolicy,
 
 /* delete the channel context */
 
-static void 
+static void
 UA_ChannelModule_Basic256_Delete_Context (void * channelContext) {
     if (channelContext != NULL) {
         Channel_Context_Basic256 * cc = (Channel_Context_Basic256 *)
                                            channelContext;
-        X509_free (cc->remoteCertificateX509);                                           
-        UA_ByteString_clear (&cc->remoteCertificate); 
+        X509_free (cc->remoteCertificateX509);
+        UA_ByteString_clear (&cc->remoteCertificate);
         UA_ByteString_clear (&cc->localSymSigningKey);
         UA_ByteString_clear (&cc->localSymEncryptingKey);
         UA_ByteString_clear (&cc->localSymIv);
         UA_ByteString_clear (&cc->remoteSymSigningKey);
         UA_ByteString_clear (&cc->remoteSymEncryptingKey);
         UA_ByteString_clear (&cc->remoteSymIv);
-        UA_LOG_INFO (cc->policyContext->logger, 
-                 UA_LOGCATEGORY_SECURITYPOLICY, 
-                 "The basic256 security policy channel with openssl is deleted.");   
+        UA_LOG_INFO (cc->policyContext->logger,
+                 UA_LOGCATEGORY_SECURITYPOLICY,
+                 "The basic256 security policy channel with openssl is deleted.");
 
-        UA_free (cc);                      
+        UA_free (cc);
     }
 }
 
-/* Compares the supplied certificate with the certificate 
- * in the endpoint context 
+/* Compares the supplied certificate with the certificate
+ * in the endpoint context
  */
 
 static UA_StatusCode
@@ -188,12 +237,12 @@ UA_Asy_Basic256_compareCertificateThumbprint (const UA_SecurityPolicy * security
 
 /* Generates a thumbprint for the specified certificate */
 
-static UA_StatusCode 
+static UA_StatusCode
 UA_Asy_Basic256_makeCertificateThumbprint (const UA_SecurityPolicy * securityPolicy,
                                   const UA_ByteString *     certificate,
                                   UA_ByteString *           thumbprint) {
-    return UA_Openssl_X509_GetCertificateThumbprint (certificate, 
-               thumbprint, false); 
+    return UA_Openssl_X509_GetCertificateThumbprint (certificate,
+               thumbprint, false);
 }
 
 static UA_StatusCode
@@ -275,7 +324,7 @@ UA_ChannelModule_Basic256_compareCertificate (const void *          channelConte
         return UA_STATUSCODE_BADINVALIDARGUMENT;
     }
 
-    const Channel_Context_Basic256 * cc = 
+    const Channel_Context_Basic256 * cc =
                      (const Channel_Context_Basic256 *) channelContext;
     return UA_OpenSSL_X509_compare (certificate, cc->remoteCertificateX509);
 }
@@ -289,10 +338,10 @@ UA_AsySig_Basic256_getRemoteSignatureSize (const void *channelContext) {
     const Channel_Context_Basic256 * cc = (const Channel_Context_Basic256 *) channelContext;
     UA_Int32 keyLen = 0;
     UA_Openssl_RSA_Public_GetKeyLength (cc->remoteCertificateX509, &keyLen);
-    return (size_t) keyLen; 
+    return (size_t) keyLen;
 }
 
-static size_t 
+static size_t
 UA_AsySig_Basic256_getLocalSignatureSize (const void *channelContext) {
     if (channelContext == NULL) {
         return UA_STATUSCODE_BADINVALIDARGUMENT;
@@ -303,11 +352,11 @@ UA_AsySig_Basic256_getLocalSignatureSize (const void *channelContext) {
     UA_Int32 keyLen = 0;
     UA_Openssl_RSA_Private_GetKeyLength (pc->localPrivateKey, &keyLen);
 
-    return (size_t) keyLen; 
+    return (size_t) keyLen;
 }
 
-static UA_StatusCode 
-UA_AsySig_Basic256_Verify (void *                    channelContext, 
+static UA_StatusCode
+UA_AsySig_Basic256_Verify (void *                    channelContext,
                            const UA_ByteString *     message,
                            const UA_ByteString *     signature) {
     if (message == NULL || signature == NULL || channelContext == NULL) {
@@ -315,17 +364,17 @@ UA_AsySig_Basic256_Verify (void *                    channelContext,
         }
 
     Channel_Context_Basic256 * cc = (Channel_Context_Basic256 *) channelContext;
-    UA_StatusCode retval = UA_OpenSSL_RSA_PKCS1_V15_SHA1_Verify (message, 
+    UA_StatusCode retval = UA_OpenSSL_RSA_PKCS1_V15_SHA1_Verify (message,
                             cc->remoteCertificateX509, signature);
     return retval;
 }
 
 static UA_StatusCode
-UA_AsySig_Basic256_Sign (void *                    channelContext, 
+UA_AsySig_Basic256_Sign (void *                    channelContext,
                          const UA_ByteString *     message,
                          UA_ByteString *           signature) {
     if (channelContext == NULL || message == NULL || signature == NULL) {
-        return UA_STATUSCODE_BADINVALIDARGUMENT; 
+        return UA_STATUSCODE_BADINVALIDARGUMENT;
     }
 
     Channel_Context_Basic256 * cc = (Channel_Context_Basic256 *) channelContext;
@@ -346,7 +395,7 @@ UA_AsymEn_Basic256_getRemotePlainTextBlockSize (const void *channelContext) {
     return (size_t) keyLen - UA_SECURITYPOLICY_BASIC256SHA1_RSAPADDING_LEN;
 }
 
-static size_t 
+static size_t
 UA_AsymEn_Basic256_getRemoteBlockSize (const void *channelContext) {
     if (channelContext == NULL) {
         return UA_STATUSCODE_BADINVALIDARGUMENT;
@@ -369,7 +418,7 @@ UA_AsymEn_Basic256_getRemoteKeyLength (const void *channelContext) {
     return (size_t) keyLen * 8;
 }
 
-static size_t 
+static size_t
 UA_AsymEn_Basic256_getLocalKeyLength (const void *channelContext) {
     if (channelContext == NULL)
         return UA_STATUSCODE_BADINVALIDARGUMENT;
@@ -379,10 +428,10 @@ UA_AsymEn_Basic256_getLocalKeyLength (const void *channelContext) {
     UA_Int32 keyLen = 0;
     UA_Openssl_RSA_Private_GetKeyLength (pc->localPrivateKey, &keyLen);
 
-    return (size_t) keyLen * 8; 
+    return (size_t) keyLen * 8;
 }
 
-static UA_StatusCode 
+static UA_StatusCode
 UA_AsymEn_Basic256_Decrypt (void *                    channelContext,
                             UA_ByteString *           data) {
     if (channelContext == NULL || data == NULL) {
@@ -390,19 +439,19 @@ UA_AsymEn_Basic256_Decrypt (void *                    channelContext,
     }
 
     Channel_Context_Basic256 * cc = (Channel_Context_Basic256 *) channelContext;
-    UA_StatusCode ret = UA_Openssl_RSA_Oaep_Decrypt (data, 
+    UA_StatusCode ret = UA_Openssl_RSA_Oaep_Decrypt (data,
                         cc->policyContext->localPrivateKey);
-    return ret;                        
+    return ret;
 }
 
 static UA_StatusCode
 UA_AsymEn_Basic256_Encrypt (void *                    channelContext,
                             UA_ByteString *           data) {
     if (channelContext == NULL || data == NULL) {
-        return UA_STATUSCODE_BADINVALIDARGUMENT; 
+        return UA_STATUSCODE_BADINVALIDARGUMENT;
     }
 
-    Channel_Context_Basic256 * cc = (Channel_Context_Basic256 *) channelContext;    
+    Channel_Context_Basic256 * cc = (Channel_Context_Basic256 *) channelContext;
     return UA_Openssl_RSA_OAEP_Encrypt (data, UA_SECURITYPOLICY_BASIC256SHA1_RSAPADDING_LEN,
                                         cc->remoteCertificateX509);
 }
@@ -423,13 +472,13 @@ UA_Sym_Basic256_generateKey(void *policyContext, const UA_ByteString *secret,
     return UA_Openssl_Random_Key_PSHA1_Derive(secret, seed, out);
 }
 
-static size_t 
+static size_t
 UA_SymEn_Basic256_getLocalKeyLength (const void *              channelContext) {
     /* 32 bytes 256 bits */
-    return UA_SECURITYPOLICY_BASIC256_SYM_ENCRYPTION_KEY_LENGTH; 
+    return UA_SECURITYPOLICY_BASIC256_SYM_ENCRYPTION_KEY_LENGTH;
 }
 
-static size_t 
+static size_t
 UA_SymEn_Basic256_getBlockSize (const void *              channelContext) {
     return UA_SECURITYPOLICY_BASIC256_SYM_ENCRYPTION_BLOCK_SIZE;
 }
@@ -437,7 +486,7 @@ UA_SymEn_Basic256_getBlockSize (const void *              channelContext) {
 static size_t
 UA_SymEn_Basic256_getRemoteKeyLength (const void * channelContext) {
     /* 32 bytes 256 bits */
-    return UA_SECURITYPOLICY_BASIC256_SYM_ENCRYPTION_KEY_LENGTH; 
+    return UA_SECURITYPOLICY_BASIC256_SYM_ENCRYPTION_KEY_LENGTH;
 }
 
 static UA_StatusCode
@@ -445,7 +494,7 @@ UA_SymEn_Basic256_Encrypt (void *                    channelContext,
                            UA_ByteString *           data) {
     if(channelContext == NULL || data == NULL)
         return UA_STATUSCODE_BADINVALIDARGUMENT;
-    
+
     Channel_Context_Basic256 * cc = (Channel_Context_Basic256 *) channelContext;
     return UA_OpenSSL_AES_256_CBC_Encrypt (&cc->localSymIv, &cc->localSymEncryptingKey, data);
 }
@@ -455,13 +504,13 @@ UA_SymEn_Basic256_Decrypt (void *                    channelContext,
                            UA_ByteString *           data) {
     if(channelContext == NULL || data == NULL)
         return UA_STATUSCODE_BADINVALIDARGUMENT;
-    Channel_Context_Basic256 * cc = (Channel_Context_Basic256 *) channelContext;    
+    Channel_Context_Basic256 * cc = (Channel_Context_Basic256 *) channelContext;
     return UA_OpenSSL_AES_256_CBC_Decrypt (&cc->remoteSymIv, &cc->remoteSymEncryptingKey, data);
 }
 
-static size_t 
+static size_t
 UA_SymSig_Basic256_getKeyLength (const void *              channelContext) {
-    return UA_SECURITYPOLICY_BASIC256_SYM_SIGNING_KEY_LENGTH; 
+    return UA_SECURITYPOLICY_BASIC256_SYM_SIGNING_KEY_LENGTH;
 }
 
 static size_t
@@ -470,26 +519,26 @@ UA_SymSig_Basic256_getSignatureSize (const void *              channelContext) {
 }
 
 static UA_StatusCode
-UA_SymSig_Basic256_Verify (void *                    channelContext, 
+UA_SymSig_Basic256_Verify (void *                    channelContext,
                            const UA_ByteString *     message,
                            const UA_ByteString *     signature) {
-    if (channelContext == NULL || 
+    if (channelContext == NULL ||
        message == NULL || signature == NULL)
         return UA_STATUSCODE_BADINVALIDARGUMENT;
-    
+
     Channel_Context_Basic256 * cc = (Channel_Context_Basic256 *) channelContext;
-    return UA_OpenSSL_HMAC_SHA1_Verify (message, 
-                                        &cc->remoteSymSigningKey, 
-                                        signature);   
+    return UA_OpenSSL_HMAC_SHA1_Verify (message,
+                                        &cc->remoteSymSigningKey,
+                                        signature);
 }
 
-static UA_StatusCode 
-UA_SymSig_Basic256_Sign (void *                    channelContext, 
+static UA_StatusCode
+UA_SymSig_Basic256_Sign (void *                    channelContext,
                          const UA_ByteString *     message,
                          UA_ByteString *           signature) {
     if (channelContext == NULL || message == NULL || signature == NULL)
         return UA_STATUSCODE_BADINVALIDARGUMENT;
-    
+
     Channel_Context_Basic256 * cc = (Channel_Context_Basic256 *) channelContext;
     return UA_OpenSSL_HMAC_SHA1_Sign (message, &cc->localSymSigningKey, signature);
 }
@@ -499,7 +548,7 @@ UA_SymSig_Basic256_Sign (void *                    channelContext,
 UA_StatusCode
 UA_SecurityPolicy_Basic256 (UA_SecurityPolicy * policy,
                             const UA_ByteString localCertificate,
-                            const UA_ByteString localPrivateKey, 
+                            const UA_ByteString localPrivateKey,
                             const UA_Logger *   logger) {
 
     UA_LOG_WARNING(logger, UA_LOGCATEGORY_SECURITYPOLICY,
@@ -507,11 +556,11 @@ UA_SecurityPolicy_Basic256 (UA_SecurityPolicy * policy,
                    "There are known attacks that break the encryption.");
 
     UA_SecurityPolicyAsymmetricModule * const asymmetricModule = &policy->asymmetricModule;
-    UA_SecurityPolicySymmetricModule * const  symmetricModule = &policy->symmetricModule; 
-    UA_SecurityPolicyChannelModule * const    channelModule = &policy->channelModule;  
-    UA_StatusCode                             retval; 
+    UA_SecurityPolicySymmetricModule * const  symmetricModule = &policy->symmetricModule;
+    UA_SecurityPolicyChannelModule * const    channelModule = &policy->channelModule;
+    UA_StatusCode                             retval;
 
-    UA_LOG_INFO (logger, UA_LOGCATEGORY_SECURITYPOLICY, 
+    UA_LOG_INFO (logger, UA_LOGCATEGORY_SECURITYPOLICY,
                  "The basic256 security policy with openssl is added.");
 
     UA_Openssl_Init ();
@@ -549,7 +598,7 @@ UA_SecurityPolicy_Basic256 (UA_SecurityPolicy * policy,
     asySigAlgorithm->uri = UA_STRING("http://www.w3.org/2000/09/xmldsig#rsa-sha1\0");
     asySigAlgorithm->getRemoteSignatureSize = UA_AsySig_Basic256_getRemoteSignatureSize;
     asySigAlgorithm->getLocalSignatureSize = UA_AsySig_Basic256_getLocalSignatureSize;
-    asySigAlgorithm->verify = UA_AsySig_Basic256_Verify;    
+    asySigAlgorithm->verify = UA_AsySig_Basic256_Verify;
     asySigAlgorithm->sign = UA_AsySig_Basic256_Sign;
     asySigAlgorithm->getLocalKeyLength = NULL;
     asySigAlgorithm->getRemoteKeyLength = NULL;
@@ -570,13 +619,13 @@ UA_SecurityPolicy_Basic256 (UA_SecurityPolicy * policy,
 
     symmetricModule->secureChannelNonceLength = 32;
     symmetricModule->generateNonce = UA_Sym_Basic256_generateNonce;
-    symmetricModule->generateKey = UA_Sym_Basic256_generateKey; 
+    symmetricModule->generateKey = UA_Sym_Basic256_generateKey;
 
     /* Symmetric encryption Algorithm */
 
     UA_SecurityPolicyEncryptionAlgorithm * symEncryptionAlgorithm =
         &symmetricModule->cryptoModule.encryptionAlgorithm;
-    symEncryptionAlgorithm->uri = UA_STRING("http://www.w3.org/2001/04/xmlenc#aes256-cbc\0");   
+    symEncryptionAlgorithm->uri = UA_STRING("http://www.w3.org/2001/04/xmlenc#aes256-cbc\0");
     symEncryptionAlgorithm->getLocalKeyLength = UA_SymEn_Basic256_getLocalKeyLength;
     symEncryptionAlgorithm->getRemoteKeyLength = UA_SymEn_Basic256_getRemoteKeyLength;
     symEncryptionAlgorithm->getRemoteBlockSize = UA_SymEn_Basic256_getBlockSize;
@@ -603,9 +652,10 @@ UA_SecurityPolicy_Basic256 (UA_SecurityPolicy * policy,
         UA_ByteString_clear (&policy->localCertificate);
         return retval;
     }
-    policy->clear = UA_Policy_Basic256_Clear_Context;    
+    policy->updateCertificateAndPrivateKey = updateCertificateAndPrivateKey_sp_basic256;
+    policy->clear = UA_Policy_Basic256_Clear_Context;
 
-    /* Use the same signature algorithm as the asymmetric component for 
+    /* Use the same signature algorithm as the asymmetric component for
        certificate signing (see standard) */
     policy->certificateSigningAlgorithm = policy->asymmetricModule.cryptoModule.signatureAlgorithm;
 
