@@ -19,6 +19,9 @@
 #include <openssl/pem.h>
 
 #include "ua_openssl_version_abstraction.h"
+#include "libc_time.h"
+
+#include <limits.h>
 
 /* Find binary substring. Taken and adjusted from
  * http://tungchingkai.blogspot.com/2011/07/binary-strstr.html */
@@ -60,62 +63,72 @@ UA_Bstrstr(const unsigned char *s1, size_t l1, const unsigned char *s2, size_t l
 }
 
 typedef struct {
-    /* 
+    /*
      * If the folders are defined, we use them to reload the certificates during
-     * runtime 
+     * runtime
      */
 
-    UA_String             trustListFolder;  
+    UA_String             trustListFolder;
     UA_String             issuerListFolder;
-    UA_String             revocationListFolder; 
+    UA_String             revocationListFolder;
+    /* Used with mbedTLS and UA_ENABLE_CERT_REJECTED_DIR option */
+    UA_String             rejectedListFolder;
 
     STACK_OF(X509) *      skIssue;
     STACK_OF(X509) *      skTrusted;
     STACK_OF(X509_CRL) *  skCrls; /* Revocation list*/
+
+    UA_CertificateVerification *cv;
 } CertContext;
 
-static UA_StatusCode 
+static UA_StatusCode
 UA_CertContext_sk_Init (CertContext * context) {
     context->skTrusted = sk_X509_new_null();
     context->skIssue = sk_X509_new_null();
-    context->skCrls = sk_X509_CRL_new_null();    
+    context->skCrls = sk_X509_CRL_new_null();
     if (context->skTrusted == NULL || context->skIssue == NULL ||
         context->skCrls == NULL) {
-        return UA_STATUSCODE_BADOUTOFMEMORY;    
+        return UA_STATUSCODE_BADOUTOFMEMORY;
     }
     return UA_STATUSCODE_GOOD;
 }
 
-static void 
+static void
 UA_CertContext_sk_free (CertContext * context) {
     sk_X509_pop_free (context->skTrusted, X509_free);
     sk_X509_pop_free (context->skIssue, X509_free);
     sk_X509_CRL_pop_free (context->skCrls, X509_CRL_free);
 }
 
-static UA_StatusCode 
-UA_CertContext_Init (CertContext * context) {
+static UA_StatusCode
+UA_CertContext_Init (CertContext * context, UA_CertificateVerification *cv) {
     (void) memset (context, 0, sizeof (CertContext));
     UA_ByteString_init (&context->trustListFolder);
     UA_ByteString_init (&context->issuerListFolder);
     UA_ByteString_init (&context->revocationListFolder);
+    UA_ByteString_init (&context->rejectedListFolder);
+
+    context->cv = cv;
+
     return UA_CertContext_sk_Init (context);
 }
 
 static void
 UA_CertificateVerification_clear (UA_CertificateVerification * cv) {
     if (cv == NULL) {
-        return ;
+        return;
     }
-    CertContext * context = (CertContext *) cv->context;    
+    CertContext * context = (CertContext *) cv->context;
     if (context == NULL) {
         return;
     }
     UA_ByteString_clear (&context->trustListFolder);
     UA_ByteString_clear (&context->issuerListFolder);
     UA_ByteString_clear (&context->revocationListFolder);
+    UA_ByteString_clear (&context->rejectedListFolder);
 
     UA_CertContext_sk_free (context);
+    context->cv = NULL;
     UA_free (context);
 
     memset(cv, 0, sizeof(UA_CertificateVerification));
@@ -125,7 +138,7 @@ static UA_StatusCode
 UA_skTrusted_Cert2X509 (const UA_ByteString *   certificateTrustList,
                         size_t                  certificateTrustListSize,
                         CertContext *           ctx) {
-    size_t                i;        
+    size_t                i;
 
     for (i = 0; i < certificateTrustListSize; i++) {
         X509 * x509 = UA_OpenSSL_LoadCertificate(&certificateTrustList[i]);
@@ -136,7 +149,7 @@ UA_skTrusted_Cert2X509 (const UA_ByteString *   certificateTrustList,
         sk_X509_push (ctx->skTrusted, x509);
     }
 
-    return UA_STATUSCODE_GOOD;                            
+    return UA_STATUSCODE_GOOD;
 }
 
 static UA_StatusCode
@@ -154,14 +167,14 @@ UA_skIssuer_Cert2X509 (const UA_ByteString *   certificateIssuerList,
         sk_X509_push (ctx->skIssue, x509);
     }
 
-    return UA_STATUSCODE_GOOD;                            
+    return UA_STATUSCODE_GOOD;
 }
 
 static UA_StatusCode
 UA_skCrls_Cert2X509 (const UA_ByteString *   certificateRevocationList,
                      size_t                  certificateRevocationListSize,
                      CertContext *           ctx) {
-    size_t                i;        
+    size_t                i;
     const unsigned char * pData;
 
     for (i = 0; i < certificateRevocationListSize; i++) {
@@ -184,10 +197,10 @@ UA_skCrls_Cert2X509 (const UA_ByteString *   certificateRevocationList,
         sk_X509_CRL_push (ctx->skCrls, crl);
     }
 
-    return UA_STATUSCODE_GOOD;                            
+    return UA_STATUSCODE_GOOD;
 }
 
-#ifdef __linux__ 
+#ifdef __linux__
 #include <dirent.h>
 
 static int UA_Certificate_Filter_der_pem (const struct dirent * entry) {
@@ -196,7 +209,7 @@ static int UA_Certificate_Filter_der_pem (const struct dirent * entry) {
 
     /* check file extension */
     const char *pszFind = strrchr(entry->d_name, '.');
-    if (pszFind == 0) 
+    if (pszFind == 0)
         return 0;
     pszFind++;
     if (strcmp (pszFind, "der") == 0 || strcmp (pszFind, "pem") == 0)
@@ -212,7 +225,7 @@ static int UA_Certificate_Filter_crl (const struct dirent * entry) {
 
     /* check file extension */
     const char *pszFind = strrchr(entry->d_name, '.');
-    if (pszFind == 0) 
+    if (pszFind == 0)
         return 0;
     pszFind++;
     if (strcmp (pszFind, "crl") == 0)
@@ -241,12 +254,12 @@ UA_BuildFullPath (const char * path,
 static UA_StatusCode
 UA_loadCertFromFile (const char *     fileName,
                      UA_ByteString *  cert) {
- 
-    FILE * fp = fopen(fileName, "rb"); 
+
+    FILE * fp = fopen(fileName, "rb");
 
     if (fp == NULL)
         return UA_STATUSCODE_BADINTERNALERROR;
-    
+
     fseek(fp, 0, SEEK_END);
     cert->length = (size_t)  ftell(fp);
     if (UA_ByteString_allocBuffer (cert, cert->length) != UA_STATUSCODE_GOOD) {
@@ -273,13 +286,13 @@ UA_ReloadCertFromFolder (CertContext * ctx) {
     int              i;
     int              numCertificates;
     char             certFile[PATH_MAX];
-    UA_ByteString    strCert; 
+    UA_ByteString    strCert;
     char             folderPath[PATH_MAX];
 
     UA_ByteString_init (&strCert);
 
     if (ctx->trustListFolder.length > 0) {
-        UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "Reloading the trust-list"); 
+        UA_LOG_INFO(ctx->cv->logging, UA_LOGCATEGORY_SERVER, "Reloading the trust-list");
 
         sk_X509_pop_free (ctx->skTrusted, X509_free);
         ctx->skTrusted = sk_X509_new_null();
@@ -287,25 +300,25 @@ UA_ReloadCertFromFolder (CertContext * ctx) {
             return UA_STATUSCODE_BADOUTOFMEMORY;
         }
 
-        (void) memcpy (folderPath, ctx->trustListFolder.data, 
+        (void) memcpy (folderPath, ctx->trustListFolder.data,
                        ctx->trustListFolder.length);
         folderPath[ctx->trustListFolder.length] = 0;
-        numCertificates = scandir(folderPath, &dirlist, 
+        numCertificates = scandir(folderPath, &dirlist,
                                   UA_Certificate_Filter_der_pem,
                                   alphasort);
         for (i = 0; i < numCertificates; i++) {
-            if (UA_BuildFullPath (folderPath, dirlist[i]->d_name, 
+            if (UA_BuildFullPath (folderPath, dirlist[i]->d_name,
                                   PATH_MAX, certFile) != UA_STATUSCODE_GOOD) {
-                continue; 
+                continue;
             }
             ret = UA_loadCertFromFile (certFile, &strCert);
             if (ret != UA_STATUSCODE_GOOD) {
-                UA_LOG_INFO (UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
-                            "Failed to load the certificate file %s", certFile); 
+                UA_LOG_INFO(ctx->cv->logging, UA_LOGCATEGORY_SERVER,
+                            "Failed to load the certificate file %s", certFile);
                 continue;  /* continue or return ? */
             }
             if (UA_skTrusted_Cert2X509 (&strCert, 1, ctx) != UA_STATUSCODE_GOOD) {
-                UA_LOG_INFO (UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
+                UA_LOG_INFO (ctx->cv->logging, UA_LOGCATEGORY_SERVER,
                             "Failed to decode the certificate file %s", certFile);
                 UA_ByteString_clear (&strCert);
                 continue;  /* continue or return ? */
@@ -315,7 +328,7 @@ UA_ReloadCertFromFolder (CertContext * ctx) {
     }
 
     if (ctx->issuerListFolder.length > 0) {
-        UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "Reloading the issuer-list");    
+        UA_LOG_INFO(ctx->cv->logging, UA_LOGCATEGORY_SERVER, "Reloading the issuer-list");
 
         sk_X509_pop_free (ctx->skIssue, X509_free);
         ctx->skIssue = sk_X509_new_null();
@@ -325,23 +338,23 @@ UA_ReloadCertFromFolder (CertContext * ctx) {
 
         memcpy (folderPath, ctx->issuerListFolder.data, ctx->issuerListFolder.length);
         folderPath[ctx->issuerListFolder.length] = 0;
-        numCertificates = scandir(folderPath, &dirlist, 
+        numCertificates = scandir(folderPath, &dirlist,
                                   UA_Certificate_Filter_der_pem,
                                   alphasort);
         for (i = 0; i < numCertificates; i++) {
-            if (UA_BuildFullPath (folderPath, dirlist[i]->d_name, 
+            if (UA_BuildFullPath (folderPath, dirlist[i]->d_name,
                                   PATH_MAX, certFile) != UA_STATUSCODE_GOOD) {
-                continue; 
+                continue;
             }
             ret = UA_loadCertFromFile (certFile, &strCert);
             if (ret != UA_STATUSCODE_GOOD) {
-                UA_LOG_INFO (UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
-                            "Failed to load the certificate file %s", certFile); 
+                UA_LOG_INFO (ctx->cv->logging, UA_LOGCATEGORY_SERVER,
+                            "Failed to load the certificate file %s", certFile);
                 continue;  /* continue or return ? */
             }
             if (UA_skIssuer_Cert2X509 (&strCert, 1, ctx) != UA_STATUSCODE_GOOD) {
-                UA_LOG_INFO (UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
-                            "Failed to decode the certificate file %s", certFile);                 
+                UA_LOG_INFO (ctx->cv->logging, UA_LOGCATEGORY_SERVER,
+                            "Failed to decode the certificate file %s", certFile);
                 UA_ByteString_clear (&strCert);
                 continue;  /* continue or return ? */
             }
@@ -350,33 +363,33 @@ UA_ReloadCertFromFolder (CertContext * ctx) {
     }
 
     if (ctx->revocationListFolder.length > 0) {
-        UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "Reloading the revocation-list");   
+        UA_LOG_INFO(ctx->cv->logging, UA_LOGCATEGORY_SERVER, "Reloading the revocation-list");
 
         sk_X509_CRL_pop_free (ctx->skCrls, X509_CRL_free);
-        ctx->skCrls = sk_X509_CRL_new_null();    
+        ctx->skCrls = sk_X509_CRL_new_null();
         if (ctx->skCrls == NULL) {
             return UA_STATUSCODE_BADOUTOFMEMORY;
         }
 
         memcpy (folderPath, ctx->revocationListFolder.data, ctx->revocationListFolder.length);
         folderPath[ctx->revocationListFolder.length] = 0;
-        numCertificates = scandir(folderPath, &dirlist, 
-                                  UA_Certificate_Filter_crl, 
+        numCertificates = scandir(folderPath, &dirlist,
+                                  UA_Certificate_Filter_crl,
                                   alphasort);
         for (i = 0; i < numCertificates; i++) {
-            if (UA_BuildFullPath (folderPath, dirlist[i]->d_name, 
+            if (UA_BuildFullPath (folderPath, dirlist[i]->d_name,
                                   PATH_MAX, certFile) != UA_STATUSCODE_GOOD) {
-                continue; 
+                continue;
             }
             ret = UA_loadCertFromFile (certFile, &strCert);
             if (ret != UA_STATUSCODE_GOOD) {
-                UA_LOG_INFO (UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
-                            "Failed to load the revocation file %s", certFile);                 
+                UA_LOG_INFO (ctx->cv->logging, UA_LOGCATEGORY_SERVER,
+                            "Failed to load the revocation file %s", certFile);
                 continue;  /* continue or return ? */
             }
             if (UA_skCrls_Cert2X509 (&strCert, 1, ctx) != UA_STATUSCODE_GOOD) {
-                UA_LOG_INFO (UA_Log_Stdout, UA_LOGCATEGORY_SERVER,
-                            "Failed to decode the revocation file %s", certFile);                                 
+                UA_LOG_INFO (ctx->cv->logging, UA_LOGCATEGORY_SERVER,
+                            "Failed to decode the revocation file %s", certFile);
                 UA_ByteString_clear (&strCert);
                 continue;  /* continue or return ? */
             }
@@ -465,53 +478,17 @@ openSSLFindNextIssuer(CertContext *ctx, STACK_OF(X509) *stack, X509 *x509, X509
             if(X509_check_issued(candidate, x509) == 0)
                 return candidate;
         }
-        /* Switch from the stack that came with the cert to the issuer list and
-         * then to the trust list. */
-        if(stack == ctx->skTrusted)
-            stack = NULL;
-        else if(stack == ctx->skIssue)
-            stack = ctx->skTrusted;
-        else
-            stack = ctx->skIssue;
+        /* Switch to search in the ctx->skIssue list */
+        stack = (stack != ctx->skIssue) ? ctx->skIssue : NULL;
     } while(stack);
     return NULL;
 }
 
-/* Is the certificate a CA? */
 static UA_Boolean
-openSSLCheckCA(X509 *cert) {
-    uint32_t flags = X509_get_extension_flags(cert);
-    /* The basic constraints must be set with the CA flag true */
-    if(!(flags & EXFLAG_CA))
-        return false;
-
-    /* The Key Usage extension must be set */
-    if(!(flags & EXFLAG_KUSAGE))
-        return false;
-
-    /* The Key Usage must include cert signing and CRL issuing */
-    uint32_t usage = X509_get_key_usage(cert);
-    if(!(usage & KU_KEY_CERT_SIGN) || !(usage & KU_CRL_SIGN))
-        return false;
-
-    return true;
-}
-
-static UA_StatusCode
 openSSLCheckRevoked(CertContext *ctx, X509 *cert) {
     const ASN1_INTEGER *sn = X509_get0_serialNumber(cert);
     const X509_NAME *in = X509_get_issuer_name(cert);
     int size = sk_X509_CRL_num(ctx->skCrls);
-
-    if(size == 0) {
-        UA_LOG_WARNING(UA_Log_Stdout, UA_LOGCATEGORY_SECURITYPOLICY,
-                       "Zero revocation lists have been loaded. "
-                       "This seems intentional - omitting the check.");
-        return UA_STATUSCODE_GOOD;
-    }
-
-    /* Loop over the crl and match the Issuer Name */
-    UA_StatusCode res = UA_STATUSCODE_BADCERTIFICATEREVOCATIONUNKNOWN;
     for(int i = 0; i < size; i++) {
         /* The crl contains a list of serial numbers from the same issuer */
         X509_CRL *crl = sk_X509_CRL_value(ctx->skCrls, i);
@@ -522,11 +499,10 @@ openSSLCheckRevoked(CertContext *ctx, X509 *cert) {
         for(int j = 0; j < rsize; j++) {
             X509_REVOKED *r = sk_X509_REVOKED_value(rs, j);
             if(ASN1_INTEGER_cmp(sn, X509_REVOKED_get0_serialNumber(r)) == 0)
-                return UA_STATUSCODE_BADCERTIFICATEREVOKED;
+                return true;
         }
-        res = UA_STATUSCODE_GOOD; /* There was at least one crl that did not revoke (so far) */
     }
-    return res;
+    return false;
 }
 
 #define UA_OPENSSL_MAX_CHAIN_LENGTH 10
@@ -545,6 +521,11 @@ openSSL_verifyChain(CertContext *ctx, STACK_OF(X509) *stack, X509 **old_issuers,
         return (depth == 0) ? UA_STATUSCODE_BADCERTIFICATETIMEINVALID :
             UA_STATUSCODE_BADCERTIFICATEISSUERTIMEINVALID;
 
+    /* Verification Step: Revocation Check */
+    if(openSSLCheckRevoked(ctx, cert))
+        return (depth == 0) ? UA_STATUSCODE_BADCERTIFICATEREVOKED :
+            UA_STATUSCODE_BADCERTIFICATEISSUERREVOKED;
+
     /* Return the most specific error code. BADCERTIFICATECHAININCOMPLETE is
      * returned only if all possible chains are incomplete. */
     X509 *issuer = NULL;
@@ -558,7 +539,7 @@ openSSL_verifyChain(CertContext *ctx, STACK_OF(X509) *stack, X509 **old_issuers,
 
         /* Verification Step: Certificate Usage
          * Can the issuer act as CA? Omit for self-signed leaf certificates. */
-        if((depth > 0 || issuer != cert) && !openSSLCheckCA(issuer)) {
+        if((depth > 0 || issuer != cert) && !X509_check_ca(issuer)) {
             ret = UA_STATUSCODE_BADCERTIFICATEISSUERUSENOTALLOWED;
             continue;
         }
@@ -576,25 +557,11 @@ openSSL_verifyChain(CertContext *ctx, STACK_OF(X509) *stack, X509 **old_issuers,
          * chain. We check whether the certificate is trusted below. This is the
          * only place where we return UA_STATUSCODE_BADCERTIFICATEUNTRUSTED.
          * This signals that the chain is complete (but can be still
-         * untrusted).
-         *
-         * Break here as we have reached the end of the chain. Omit the
-         * Revocation Check for self-signed certificates. */
+         * untrusted). */
         if(cert == issuer || X509_cmp(cert, issuer) == 0) {
             ret = UA_STATUSCODE_BADCERTIFICATEUNTRUSTED;
-            break;
-        }
-
-        /* Verification Step: Revocation Check */
-        ret = openSSLCheckRevoked(ctx, cert);
-        if(depth > 0) {
-            if(ret == UA_STATUSCODE_BADCERTIFICATEREVOKED)
-                ret = UA_STATUSCODE_BADCERTIFICATEISSUERREVOKED;
-            if(ret == UA_STATUSCODE_BADCERTIFICATEREVOCATIONUNKNOWN)
-                ret = UA_STATUSCODE_BADCERTIFICATEISSUERREVOCATIONUNKNOWN;
-        }
-        if(ret != UA_STATUSCODE_GOOD)
             continue;
+        }
 
         /* Detect (endless) loops of issuers. The last one can be skipped by the
          * check for self-signed just before. */
@@ -623,13 +590,13 @@ openSSL_verifyChain(CertContext *ctx, STACK_OF(X509) *stack, X509 **old_issuers,
 /* This follows Part 6, 6.1.3 Determining if a Certificate is trusted.
  * It defines a sequence of steps for certificate verification. */
 static UA_StatusCode
-UA_CertificateVerification_Verify(void *verificationContext,
+UA_CertificateVerification_Verify(const UA_CertificateVerification *cv,
                                   const UA_ByteString *certificate) {
-    if(!verificationContext || !certificate)
+    if(!cv || !certificate)
         return UA_STATUSCODE_BADINTERNALERROR;
 
     UA_StatusCode ret = UA_STATUSCODE_GOOD;
-    CertContext *ctx = (CertContext *)verificationContext;
+    CertContext *ctx = (CertContext *)cv->context;
 
 #ifdef __linux__ 
     ret = UA_ReloadCertFromFolder(ctx);
@@ -650,7 +617,7 @@ UA_CertificateVerification_Verify(void *verificationContext,
      * Refer the test case CTT/Security/Security Certificate Validation/029.js
      * for more details. */
     X509 *leaf = sk_X509_value(stack, 0);
-    if(openSSLCheckCA(leaf)) {
+    if(X509_check_ca(leaf)) {
         sk_X509_pop_free(stack, X509_free);
         return UA_STATUSCODE_BADCERTIFICATEUSENOTALLOWED;
     }
@@ -670,32 +637,42 @@ UA_CertificateVerification_Verify(void *verificationContext,
 }
 
 static UA_StatusCode
-UA_CertificateVerification_VerifyApplicationURI (void *                verificationContext,
+UA_CertificateVerification_VerifyApplicationURI (const UA_CertificateVerification *cv,
                                                  const UA_ByteString * certificate,
                                                  const UA_String *     applicationURI) {
-    (void) verificationContext;
-
     const unsigned char * pData;
     X509 *                certificateX509;
     UA_String             subjectURI = UA_STRING_NULL;
     GENERAL_NAMES *       pNames;
     int                   i;
     UA_StatusCode         ret;
+    CertContext *         ctx;
+
+    if (cv == NULL) {
+        return UA_STATUSCODE_BADINTERNALERROR;
+    }
+    ctx = (CertContext *)cv->context;
+    if (ctx == NULL) {
+        return UA_STATUSCODE_BADINTERNALERROR;
+    }
 
     pData = certificate->data;
     if (pData == NULL) {
+        UA_LOG_ERROR(cv->logging, UA_LOGCATEGORY_USERLAND, "Error Empty Certificate");
         return UA_STATUSCODE_BADSECURITYCHECKSFAILED;
     }
 
     certificateX509 = UA_OpenSSL_LoadCertificate(certificate);
     if (certificateX509 == NULL) {
+        UA_LOG_ERROR(cv->logging, UA_LOGCATEGORY_USERLAND, "Error loading X509 Certificate");
         return UA_STATUSCODE_BADSECURITYCHECKSFAILED;
     }
 
-    pNames = (GENERAL_NAMES *) X509_get_ext_d2i(certificateX509, NID_subject_alt_name, 
+    pNames = (GENERAL_NAMES *) X509_get_ext_d2i(certificateX509, NID_subject_alt_name,
                                                 NULL, NULL);
     if (pNames == NULL) {
         X509_free (certificateX509);
+        UA_LOG_ERROR(cv->logging, UA_LOGCATEGORY_USERLAND, "Error processing X509 Certificate");
         return UA_STATUSCODE_BADSECURITYCHECKSFAILED;
     }
     
@@ -707,6 +684,7 @@ UA_CertificateVerification_VerifyApplicationURI (void *                verificat
              subjectURI.length = (size_t) (value->d.ia5->length);
              subjectURI.data = (UA_Byte *) UA_malloc (subjectURI.length);
              if (subjectURI.data == NULL) {
+                 UA_LOG_ERROR(cv->logging, UA_LOGCATEGORY_USERLAND, "Error Empty subjectURI");
                  X509_free (certificateX509);
                  sk_GENERAL_NAME_pop_free(pNames, GENERAL_NAME_free);
                  return UA_STATUSCODE_BADSECURITYCHECKSFAILED;
@@ -720,6 +698,7 @@ UA_CertificateVerification_VerifyApplicationURI (void *                verificat
     ret = UA_STATUSCODE_GOOD;
     if (UA_Bstrstr (subjectURI.data, subjectURI.length,
                     applicationURI->data, applicationURI->length) == NULL) {
+        UA_LOG_ERROR(cv->logging, UA_LOGCATEGORY_USERLAND, "Empty comparing subjectURI and applicationURI");
         ret = UA_STATUSCODE_BADCERTIFICATEURIINVALID;
     }
 
@@ -729,6 +708,53 @@ UA_CertificateVerification_VerifyApplicationURI (void *                verificat
     return ret;
 }
 
+#ifdef UA_ENABLE_ENCRYPTION_OPENSSL
+static UA_StatusCode
+UA_GetCertificate_ExpirationDate(UA_DateTime *expiryDateTime, 
+                                 UA_ByteString *certificate) {
+    const unsigned char *pData = certificate->data;
+    X509 * x509 = d2i_X509 (NULL, &pData, (long)certificate->length);
+    if (x509 == NULL) {
+        return UA_STATUSCODE_BADINTERNALERROR;
+    }
+
+    /* Get the certificate Expiry date */
+    ASN1_TIME *not_after = X509_get_notAfter(x509);
+
+    struct tm dtTime;
+    ASN1_TIME_to_tm(not_after, &dtTime);
+    X509_free(x509);
+
+    struct mytm dateTime;
+    memset(&dateTime, 0, sizeof(struct mytm));
+    dateTime.tm_year = dtTime.tm_year;
+    dateTime.tm_mon = dtTime.tm_mon;
+    dateTime.tm_mday = dtTime.tm_mday;
+    dateTime.tm_hour = dtTime.tm_hour;
+    dateTime.tm_min = dtTime.tm_min;
+    dateTime.tm_sec = dtTime.tm_sec;
+
+    long long sec_epoch = __tm_to_secs(&dateTime);
+    *expiryDateTime = UA_DATETIME_UNIX_EPOCH;
+    *expiryDateTime += sec_epoch * UA_DATETIME_SEC;
+    return UA_STATUSCODE_GOOD;
+}
+#endif
+
+static UA_StatusCode
+UA_GetCertificate_SubjectName(UA_String *subjectName,
+                              UA_ByteString *certificate) {
+    const unsigned char *pData = certificate->data;
+    X509 *x509 = d2i_X509 (NULL, &pData, (long)certificate->length);
+    if(!x509)
+        return UA_STATUSCODE_BADINTERNALERROR;
+    X509_NAME *sn = X509_get_subject_name(x509);
+    char buf[1024];
+    *subjectName = UA_STRING_ALLOC(X509_NAME_oneline(sn, buf, 1024));
+    X509_free(x509);
+    return UA_STATUSCODE_GOOD;
+}
+
 /* main entry */
 
 UA_StatusCode
@@ -744,20 +770,32 @@ UA_CertificateVerification_Trustlist(UA_CertificateVerification * cv,
     if (cv == NULL) {
         return UA_STATUSCODE_BADINTERNALERROR;
     }
+    if (cv->logging == NULL) {
+        return UA_STATUSCODE_BADINTERNALERROR;
+    }
+
+    /* Clear if the plugin is already initialized */
+    if(cv->clear)
+        cv->clear(cv);
 
     CertContext * context = (CertContext *) UA_malloc (sizeof (CertContext));
     if (context == NULL) {
         return UA_STATUSCODE_BADOUTOFMEMORY;
     }
-    ret = UA_CertContext_Init (context);
-    if (ret != UA_STATUSCODE_GOOD) {
-        return ret;
-    }
 
     cv->verifyApplicationURI = UA_CertificateVerification_VerifyApplicationURI;
     cv->clear = UA_CertificateVerification_clear;
     cv->context = context;
     cv->verifyCertificate = UA_CertificateVerification_Verify;
+#ifdef UA_ENABLE_ENCRYPTION_OPENSSL
+    cv->getExpirationDate     = UA_GetCertificate_ExpirationDate;
+#endif
+    cv->getSubjectName = UA_GetCertificate_SubjectName;
+
+    ret = UA_CertContext_Init (context, cv);
+    if (ret != UA_STATUSCODE_GOOD) {
+        return ret;
+    }
     
     if (certificateTrustListSize > 0) {
         if (UA_skTrusted_Cert2X509 (certificateTrustList, certificateTrustListSize,
@@ -770,7 +808,7 @@ UA_CertificateVerification_Trustlist(UA_CertificateVerification * cv,
     if (certificateIssuerListSize > 0) {
         if (UA_skIssuer_Cert2X509 (certificateIssuerList, certificateIssuerListSize,
                                   context) != UA_STATUSCODE_GOOD) {
-            ret = UA_STATUSCODE_BADINTERNALERROR;                                       
+            ret = UA_STATUSCODE_BADINTERNALERROR;
             goto errout;
         }
     }
@@ -778,9 +816,9 @@ UA_CertificateVerification_Trustlist(UA_CertificateVerification * cv,
     if (certificateRevocationListSize > 0) {
         if (UA_skCrls_Cert2X509 (certificateRevocationList, certificateRevocationListSize,
                                   context) != UA_STATUSCODE_GOOD) {
-            ret = UA_STATUSCODE_BADINTERNALERROR; 
+            ret = UA_STATUSCODE_BADINTERNALERROR;
             goto errout;
-        }        
+        }
     }
 
     return UA_STATUSCODE_GOOD;
@@ -799,22 +837,30 @@ UA_CertificateVerification_CertFolders(UA_CertificateVerification * cv,
     UA_StatusCode ret;
     if (cv == NULL) {
         return UA_STATUSCODE_BADINTERNALERROR;
-    }   
+    }
+    if (cv->logging == NULL) {
+        return UA_STATUSCODE_BADINTERNALERROR;
+    }
+
+    /* Clear if the plugin is already initialized */
+    if(cv->clear)
+        cv->clear(cv);
 
     CertContext * context = (CertContext *) UA_malloc (sizeof (CertContext));
     if (context == NULL) {
         return UA_STATUSCODE_BADOUTOFMEMORY;
     }
-    ret = UA_CertContext_Init (context);
-    if (ret != UA_STATUSCODE_GOOD) {
-        return ret;
-    }
 
     cv->verifyApplicationURI = UA_CertificateVerification_VerifyApplicationURI;
     cv->clear = UA_CertificateVerification_clear;
     cv->context = context;
     cv->verifyCertificate = UA_CertificateVerification_Verify;
 
+    ret = UA_CertContext_Init (context, cv);
+    if (ret != UA_STATUSCODE_GOOD) {
+        return ret;
+    }
+
     /* Only set the folder paths. They will be reloaded during runtime. */
 
     context->trustListFolder = UA_STRING_ALLOC(trustListFolder);
@@ -825,4 +871,59 @@ UA_CertificateVerification_CertFolders(UA_CertificateVerification * cv,
 }
 #endif
 
+static int
+privateKeyPasswordCallback(char *buf, int size, int rwflag, void *userdata) {
+    (void) rwflag;
+    UA_ByteString *pw = (UA_ByteString*)userdata;
+    if(pw->length <= (size_t)size)
+        memcpy(buf, pw->data, pw->length);
+    return (int)pw->length;
+}
+
+UA_StatusCode
+UA_PKI_decryptPrivateKey(const UA_ByteString privateKey,
+                         const UA_ByteString password,
+                         UA_ByteString *outDerKey) {
+    if(!outDerKey)
+        return UA_STATUSCODE_BADINTERNALERROR;
+
+    if (privateKey.length == 0) {
+        *outDerKey = UA_BYTESTRING_NULL;
+        return UA_STATUSCODE_BADINVALIDARGUMENT;
+    }
+
+    /* Already in DER format -> return verbatim */
+    if(privateKey.length > 1 && privateKey.data[0] == 0x30 && privateKey.data[1] == 0x82)
+        return UA_ByteString_copy(&privateKey, outDerKey);
+
+    /* Decrypt */
+    BIO *bio = BIO_new_mem_buf((void*)privateKey.data, (int)privateKey.length);
+    EVP_PKEY *pkey = PEM_read_bio_PrivateKey(bio, NULL,
+                                             privateKeyPasswordCallback,
+                                             (void*)(uintptr_t)&password);
+    BIO_free(bio);
+    if(!pkey)
+        return UA_STATUSCODE_BADSECURITYCHECKSFAILED;
+
+    /* Write DER encoded, allocates the new memory */
+    unsigned char *data = NULL;
+    const int numBytes = i2d_PrivateKey(pkey, &data);
+    EVP_PKEY_free(pkey);
+    if(!data)
+        return UA_STATUSCODE_BADOUTOFMEMORY;
+
+    /* Copy to the data to outDerKey
+     * Passing the data pointer directly causes a heap corruption on Windows
+     * when outDerKey is cleared.
+     */
+    UA_ByteString temp = UA_BYTESTRING_NULL;
+    temp.data = data;
+    temp.length = (size_t)numBytes;
+    const UA_StatusCode success = UA_ByteString_copy(&temp, outDerKey);
+    /* OPENSSL_clear_free() is not supported by the LibreSSL version in the CI */
+    OPENSSL_cleanse(data, numBytes);
+    OPENSSL_free(data);
+    return success;
+}
+
 #endif  /* end of defined(UA_ENABLE_ENCRYPTION_OPENSSL) || defined(UA_ENABLE_ENCRYPTION_LIBRESSL) */
