@@ -1,6 +1,6 @@
 /* This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/. 
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
  *    Copyright 2015-2016 (c) Sten Grüner
  *    Copyright 2015-2017 (c) Fraunhofer IOSB (Author: Julius Pfrommer)
@@ -20,7 +20,8 @@
 
 #include "open62541_queue.h"
 #include "ua_securechannel.h"
-#include "ua_timer.h"
+#include "ua_util_internal.h"
+#include "ziptree.h"
 
 _UA_BEGIN_DECLS
 
@@ -28,15 +29,13 @@ _UA_BEGIN_DECLS
 /* Subscriptions Handling */
 /**************************/
 
-#ifdef UA_ENABLE_SUBSCRIPTIONS
-
 typedef struct UA_Client_NotificationsAckNumber {
     LIST_ENTRY(UA_Client_NotificationsAckNumber) listEntry;
     UA_SubscriptionAcknowledgement subAck;
 } UA_Client_NotificationsAckNumber;
 
 typedef struct UA_Client_MonitoredItem {
-    LIST_ENTRY(UA_Client_MonitoredItem) listEntry;
+    ZIP_ENTRY(UA_Client_MonitoredItem) zipfields;
     UA_UInt32 monitoredItemId;
     UA_UInt32 clientHandle;
     void *context;
@@ -48,6 +47,9 @@ typedef struct UA_Client_MonitoredItem {
     UA_Boolean isEventMonitoredItem; /* Otherwise a DataChange MoniitoredItem */
 } UA_Client_MonitoredItem;
 
+ZIP_HEAD(MonitorItemsTree, UA_Client_MonitoredItem);
+typedef struct MonitorItemsTree MonitorItemsTree;
+
 typedef struct UA_Client_Subscription {
     LIST_ENTRY(UA_Client_Subscription) listEntry;
     UA_UInt32 subscriptionId;
@@ -58,23 +60,21 @@ typedef struct UA_Client_Subscription {
     UA_Client_DeleteSubscriptionCallback deleteCallback;
     UA_UInt32 sequenceNumber;
     UA_DateTime lastActivity;
-    LIST_HEAD(, UA_Client_MonitoredItem) monitoredItems;
+    MonitorItemsTree monitoredItems;
 } UA_Client_Subscription;
 
 void
-UA_Client_Subscriptions_clean(UA_Client *client);
+__Client_Subscriptions_clean(UA_Client *client);
 
 /* Exposed for fuzzing */
 UA_StatusCode
-UA_Client_preparePublishRequest(UA_Client *client, UA_PublishRequest *request);
+__Client_preparePublishRequest(UA_Client *client, UA_PublishRequest *request);
 
 void
-UA_Client_Subscriptions_backgroundPublish(UA_Client *client);
+__Client_Subscriptions_backgroundPublish(UA_Client *client);
 
 void
-UA_Client_Subscriptions_backgroundPublishInactivityCheck(UA_Client *client);
-
-#endif /* UA_ENABLE_SUBSCRIPTIONS */
+__Client_Subscriptions_backgroundPublishInactivityCheck(UA_Client *client);
 
 /**********/
 /* Client */
@@ -82,23 +82,23 @@ UA_Client_Subscriptions_backgroundPublishInactivityCheck(UA_Client *client);
 
 typedef struct AsyncServiceCall {
     LIST_ENTRY(AsyncServiceCall) pointers;
-    UA_UInt32 requestId;
+    UA_UInt32 requestId;     /* Unique id */
+    UA_UInt32 requestHandle; /* Potentially non-unique if manually defined in
+                              * the request header*/
     UA_ClientAsyncServiceCallback callback;
     const UA_DataType *responseType;
     void *userdata;
     UA_DateTime start;
     UA_UInt32 timeout;
-    void *responsedata;
+    UA_Response *syncResponse; /* If non-null, then this is the synchronous
+                                * response to be filled. Set back to null to
+                                * indicate that the response was filled. */
 } AsyncServiceCall;
 
-void
-UA_Client_AsyncService_cancel(UA_Client *client, AsyncServiceCall *ac,
-                              UA_StatusCode statusCode);
-
 typedef LIST_HEAD(UA_AsyncServiceList, AsyncServiceCall) UA_AsyncServiceList;
 
 void
-UA_Client_AsyncService_removeAll(UA_Client *client, UA_StatusCode statusCode);
+__Client_AsyncService_removeAll(UA_Client *client, UA_StatusCode statusCode);
 
 typedef struct CustomCallback {
     UA_UInt32 callbackId;
@@ -111,7 +111,9 @@ typedef struct CustomCallback {
 
 struct UA_Client {
     UA_ClientConfig config;
-    UA_Timer timer;
+
+    /* Callback ID to remove it from the EventLoop */
+    UA_UInt64 houseKeepingCallbackId;
 
     /* Overall connection status */
     UA_StatusCode connectStatus;
@@ -123,26 +125,32 @@ struct UA_Client {
 
     UA_Boolean findServersHandshake;   /* Ongoing FindServers */
     UA_Boolean endpointsHandshake;     /* Ongoing GetEndpoints */
-    UA_Boolean noSession;              /* Don't open a session */
 
-    /* Connection */
-    UA_Connection connection;
-    UA_String endpointUrl;  /* Used to extract address and port */
-    UA_String discoveryUrl; /* The discoveryUrl (also used to signal which
-                               application we want to connect to in the HEL/ACK
-                               handshake). */
+    /* The discoveryUrl can be different from the EndpointUrl in the client
+     * configuration. The EndpointUrl is used to connect initially, then the
+     * DiscoveryUrl is selected via FindServers. This triggers a reconnect if
+     * EndpointUrl != DiscoveryUrl. */
+    UA_String discoveryUrl;
+
+    /* Contains the Server description, etc. */
+    UA_EndpointDescription endpoint;
 
     /* SecureChannel */
     UA_SecureChannel channel;
-    UA_UInt32 requestId;
+    UA_UInt32 requestId; /* Unique, internally defined for each request */
     UA_DateTime nextChannelRenewal;
 
+    /* Reverse connect (listen) connections */
+    UA_ConnectionManager *reverseConnectionCM;
+    uintptr_t reverseConnectionIds[16];
+
     /* Session */
     UA_SessionState sessionState;
     UA_NodeId authenticationToken;
-    UA_UInt32 requestHandle;
-    UA_ByteString remoteNonce;
-    UA_ByteString localNonce;
+    UA_UInt32 requestHandle; /* Unique handles >100,000 are generated if the
+                              * request header contains a zero-handle. */
+    UA_ByteString serverSessionNonce;
+    UA_ByteString clientSessionNonce;
 
     /* Connectivity check */
     UA_DateTime lastConnectivityCheck;
@@ -152,26 +160,55 @@ struct UA_Client {
     UA_AsyncServiceList asyncServiceCalls;
 
     /* Subscriptions */
-#ifdef UA_ENABLE_SUBSCRIPTIONS
     LIST_HEAD(, UA_Client_NotificationsAckNumber) pendingNotificationsAcks;
     LIST_HEAD(, UA_Client_Subscription) subscriptions;
     UA_UInt32 monitoredItemHandles;
     UA_UInt16 currentlyOutStandingPublishRequests;
+
+    /* Internal locking for thread-safety. Methods starting with UA_Client_ that
+     * are marked with UA_THREADSAFE take the lock. The lock is released before
+     * dropping into the EventLoop and before calling user-defined callbacks.
+     * That way user-defined callbacks can themselves call thread-safe client
+     * methods. */
+#if UA_MULTITHREADING >= 100
+    UA_Lock clientMutex;
 #endif
 };
 
-UA_StatusCode connectSync(UA_Client *client); /* Only exposed for unit tests */
+UA_StatusCode
+__Client_AsyncService(UA_Client *client, const void *request,
+                      const UA_DataType *requestType,
+                      UA_ClientAsyncServiceCallback callback,
+                      const UA_DataType *responseType,
+                      void *userdata, UA_UInt32 *requestId);
+
+void
+__Client_Service(UA_Client *client, const void *request,
+                 const UA_DataType *requestType, void *response,
+                 const UA_DataType *responseType);
+
+UA_StatusCode
+__UA_Client_startup(UA_Client *client);
+
+UA_StatusCode
+__Client_renewSecureChannel(UA_Client *client);
+
+UA_StatusCode
+processServiceResponse(void *application, UA_SecureChannel *channel,
+                       UA_MessageType messageType, UA_UInt32 requestId,
+                       UA_ByteString *message);
+
+UA_StatusCode connectInternal(UA_Client *client, UA_Boolean async);
+UA_StatusCode connectSecureChannel(UA_Client *client, const char *endpointUrl);
+UA_Boolean isFullyConnected(UA_Client *client);
+void connectSync(UA_Client *client);
 void notifyClientState(UA_Client *client);
+void processRHEMessage(UA_Client *client, const UA_ByteString *chunk);
 void processERRResponse(UA_Client *client, const UA_ByteString *chunk);
 void processACKResponse(UA_Client *client, const UA_ByteString *chunk);
 void processOPNResponse(UA_Client *client, const UA_ByteString *message);
 void closeSecureChannel(UA_Client *client);
-
-UA_StatusCode
-connectIterate(UA_Client *client, UA_UInt32 timeout);
-
-UA_StatusCode
-receiveResponseAsync(UA_Client *client, UA_UInt32 timeout);
+void cleanupSession(UA_Client *client);
 
 void
 Client_warnEndpointsResult(UA_Client *client,
