@@ -13,7 +13,37 @@
 
 #include "ua_server_internal.h"
 #include "ua_types_encoding_binary.h"
-#include "aa_tree.h"
+
+/*********************/
+/* ReferenceType Set */
+/*********************/
+
+#define UA_REFTYPES_ALL_MASK (~(UA_UInt32)0)
+#define UA_REFTYPES_ALL_MASK2 UA_REFTYPES_ALL_MASK, UA_REFTYPES_ALL_MASK
+#define UA_REFTYPES_ALL_MASK4 UA_REFTYPES_ALL_MASK2, UA_REFTYPES_ALL_MASK2
+#if (UA_REFERENCETYPESET_MAX) / 32 > 8
+# error Adjust macros to support than 256 reference types
+#elif (UA_REFERENCETYPESET_MAX) / 32 == 8
+# define UA_REFTYPES_ALL_ARRAY UA_REFTYPES_ALL_MASK4, UA_REFTYPES_ALL_MASK4
+#elif (UA_REFERENCETYPESET_MAX) / 32 == 7
+# define UA_REFTYPES_ALL_ARRAY                                          \
+    UA_REFTYPES_ALL_MASK4, UA_REFTYPES_ALL_MASK2, UA_REFTYPES_ALL_MASK
+#elif (UA_REFERENCETYPESET_MAX) / 32 == 6
+# define UA_REFTYPES_ALL_ARRAY UA_REFTYPES_ALL_MASK4, UA_REFTYPES_ALL_MASK2
+#elif (UA_REFERENCETYPESET_MAX) / 32 == 5
+# define UA_REFTYPES_ALL_ARRAY UA_REFTYPES_ALL_MASK4, UA_REFTYPES_ALL_MASK
+#elif (UA_REFERENCETYPESET_MAX) / 32 == 4
+# define UA_REFTYPES_ALL_ARRAY UA_REFTYPES_ALL_MASK4
+#elif (UA_REFERENCETYPESET_MAX) / 32 == 3
+# define UA_REFTYPES_ALL_ARRAY UA_REFTYPES_ALL_MASK2, UA_REFTYPES_ALL_MASK
+#elif (UA_REFERENCETYPESET_MAX) / 32 == 2
+# define UA_REFTYPES_ALL_ARRAY UA_REFTYPES_ALL_MASK2
+#else
+# define UA_REFTYPES_ALL_ARRAY UA_REFTYPES_ALL_MASK
+#endif
+
+const UA_ReferenceTypeSet UA_REFERENCETYPESET_NONE = {{0}};
+const UA_ReferenceTypeSet UA_REFERENCETYPESET_ALL = {{UA_REFTYPES_ALL_ARRAY}};
 
 /*****************/
 /* Node Pointers */
@@ -225,72 +255,57 @@ static UA_StatusCode
 addReferenceTarget(UA_NodeReferenceKind *refs, UA_NodePointer target,
                    UA_UInt32 targetNameHash);
 
-static enum aa_cmp
+static UA_StatusCode
+addReferenceTargetToTree(UA_NodeReferenceKind *rk, UA_NodePointer targetId,
+                         UA_UInt32 targetIdHash, UA_UInt32 targetNameHash);
+
+enum ZIP_CMP
 cmpRefTargetId(const void *a, const void *b) {
     const UA_ReferenceTargetTreeElem *aa = (const UA_ReferenceTargetTreeElem*)a;
     const UA_ReferenceTargetTreeElem *bb = (const UA_ReferenceTargetTreeElem*)b;
     if(aa->targetIdHash < bb->targetIdHash)
-        return AA_CMP_LESS;
+        return ZIP_CMP_LESS;
     if(aa->targetIdHash > bb->targetIdHash)
-        return AA_CMP_MORE;
-    return (enum aa_cmp)UA_NodePointer_order(aa->target.targetId,
-                                             bb->target.targetId);
+        return ZIP_CMP_MORE;
+    return (enum ZIP_CMP)UA_NodePointer_order(aa->target.targetId,
+                                              bb->target.targetId);
 }
 
-static enum aa_cmp
+enum ZIP_CMP
 cmpRefTargetName(const void *a, const void *b) {
-    const UA_UInt32 *nameHashA = (const UA_UInt32*)a;
-    const UA_UInt32 *nameHashB = (const UA_UInt32*)b;
-    if(*nameHashA < *nameHashB)
-        return AA_CMP_LESS;
-    if(*nameHashA > *nameHashB)
-        return AA_CMP_MORE;
-    return AA_CMP_EQ;
+    const UA_ReferenceTargetTreeElem *aa = (const UA_ReferenceTargetTreeElem*)a;
+    const UA_ReferenceTargetTreeElem *bb = (const UA_ReferenceTargetTreeElem*)b;
+    if(aa->target.targetNameHash == bb->target.targetNameHash)
+        return ZIP_CMP_EQ;
+    return (aa->target.targetNameHash < bb->target.targetNameHash) ?
+        ZIP_CMP_LESS : ZIP_CMP_MORE;
 }
 
-/* Reusable binary search tree "heads". Just switch out the root pointer. */
-static const struct aa_head refIdTree =
-    { NULL, cmpRefTargetId, offsetof(UA_ReferenceTargetTreeElem, idTreeEntry), 0 };
-const struct aa_head refNameTree =
-    { NULL, cmpRefTargetName, offsetof(UA_ReferenceTargetTreeElem, nameTreeEntry),
-      offsetof(UA_ReferenceTarget, targetNameHash) };
-
-const UA_ReferenceTarget *
-UA_NodeReferenceKind_iterate(const UA_NodeReferenceKind *rk,
-                             const UA_ReferenceTarget *prev) {
-    /* Return from the tree */
-    if(rk->hasRefTree) {
-        const struct aa_head _refIdTree =
-            { rk->targets.tree.idTreeRoot, cmpRefTargetId,
-              offsetof(UA_ReferenceTargetTreeElem, idTreeEntry), 0 };
-        if(prev == NULL)
-            return (const UA_ReferenceTarget*)aa_min(&_refIdTree);
-        return (const UA_ReferenceTarget*)aa_next(&_refIdTree, prev);
-    }
-    if(prev == NULL) /* Return start of the array */
-        return rk->targets.array;
-    if(prev + 1 >= &rk->targets.array[rk->targetsSize])
-        return NULL; /* End of the array */
-    return prev + 1; /* Next element in the array */
-}
-
-/* Also deletes the elements of the tree */
+/* Move to the array in-order, also deletes the tree elements */
 static void
 moveTreeToArray(UA_ReferenceTarget *array, size_t *pos,
-                struct aa_entry *entry) {
-    if(!entry)
+                UA_ReferenceTargetTreeElem *elem) {
+    if(!elem)
         return;
-    UA_ReferenceTargetTreeElem *elem = (UA_ReferenceTargetTreeElem*)
-        ((uintptr_t)entry - offsetof(UA_ReferenceTargetTreeElem, idTreeEntry));
     moveTreeToArray(array, pos, elem->idTreeEntry.left);
-    moveTreeToArray(array, pos, elem->idTreeEntry.right);
     array[*pos] = elem->target;
     (*pos)++;
+    moveTreeToArray(array, pos, elem->idTreeEntry.right);
     UA_free(elem);
 }
 
+static void *
+removeTreeEntry(void *context, UA_ReferenceTargetTreeElem *elem) {
+    (void)context;
+    UA_NodePointer_clear(&elem->target.targetId);
+    UA_free(elem);
+    return NULL;
+}
+
 UA_StatusCode
 UA_NodeReferenceKind_switch(UA_NodeReferenceKind *rk) {
+    UA_assert(rk->targetsSize > 0);
+
     if(rk->hasRefTree) {
         /* From tree to array */
         UA_ReferenceTarget *array = (UA_ReferenceTarget*)
@@ -298,7 +313,7 @@ UA_NodeReferenceKind_switch(UA_NodeReferenceKind *rk) {
         if(!array)
             return UA_STATUSCODE_BADOUTOFMEMORY;
         size_t pos = 0;
-        moveTreeToArray(array, &pos, rk->targets.tree.idTreeRoot);
+        moveTreeToArray(array, &pos, rk->targets.tree.idRoot);
         rk->targets.array = array;
         rk->hasRefTree = false;
         return UA_STATUSCODE_GOOD;
@@ -307,23 +322,17 @@ UA_NodeReferenceKind_switch(UA_NodeReferenceKind *rk) {
     /* From array to tree */
     UA_NodeReferenceKind newRk = *rk;
     newRk.hasRefTree = true;
-    newRk.targets.tree.idTreeRoot = NULL;
-    newRk.targets.tree.nameTreeRoot = NULL;
+    newRk.targets.tree.idRoot = NULL;
+    newRk.targets.tree.nameRoot = NULL;
+    newRk.targetsSize = 0;
     for(size_t i = 0; i < rk->targetsSize; i++) {
         UA_StatusCode res =
             addReferenceTarget(&newRk, rk->targets.array[i].targetId,
                                rk->targets.array[i].targetNameHash);
         if(res != UA_STATUSCODE_GOOD) {
-            struct aa_head _refIdTree = refIdTree;
-            _refIdTree.root = newRk.targets.tree.idTreeRoot;
-            while(_refIdTree.root) {
-                UA_ReferenceTargetTreeElem *elem = (UA_ReferenceTargetTreeElem*)
-                    ((uintptr_t)_refIdTree.root -
-                     offsetof(UA_ReferenceTargetTreeElem, idTreeEntry));
-                aa_remove(&_refIdTree, elem);
-                UA_NodePointer_clear(&elem->target.targetId);
-                UA_free(elem);
-            }
+            ZIP_ITER(UA_ReferenceIdTree,
+                     (UA_ReferenceIdTree*)&newRk.targets.tree.idRoot,
+                     removeTreeEntry, NULL);
             return res;
         }
     }
@@ -334,36 +343,44 @@ UA_NodeReferenceKind_switch(UA_NodeReferenceKind *rk) {
     return UA_STATUSCODE_GOOD;
 }
 
-const UA_ReferenceTarget *
-UA_NodeReferenceKind_findTarget(const UA_NodeReferenceKind *rk,
-                                const UA_ExpandedNodeId *targetId) {
-    UA_NodePointer targetP = UA_NodePointer_fromExpandedNodeId(targetId);
-
-    /* Return from the tree */
-    if(rk->hasRefTree) {
-        UA_ReferenceTargetTreeElem tmpTarget;
-        tmpTarget.target.targetId = targetP;
-        tmpTarget.targetIdHash = UA_ExpandedNodeId_hash(targetId);
-        const struct aa_head _refIdTree =
-            { rk->targets.tree.idTreeRoot, cmpRefTargetId,
-              offsetof(UA_ReferenceTargetTreeElem, idTreeEntry), 0 };
-        return (const UA_ReferenceTarget*)aa_find(&_refIdTree, &tmpTarget);
-    }
-
-    /* Return from the array */
+void *
+UA_NodeReferenceKind_iterate(UA_NodeReferenceKind *rk,
+                             UA_NodeReferenceKind_iterateCallback callback,
+                             void *context) {
+    if(rk->hasRefTree)
+        return ZIP_ITER(UA_ReferenceIdTree,
+                        (UA_ReferenceIdTree*)&rk->targets.tree.idRoot,
+                        (UA_ReferenceIdTree_cb)callback, context);
     for(size_t i = 0; i < rk->targetsSize; i++) {
-        if(UA_NodePointer_equal(targetP, rk->targets.array[i].targetId))
-            return &rk->targets.array[i];
+        void *res = callback(context, &rk->targets.array[i]);
+        if(res)
+            return res;
     }
     return NULL;
 }
 
-const UA_Node *
-UA_NODESTORE_GETFROMREF(UA_Server *server, UA_NodePointer target) {
-    if(!UA_NodePointer_isLocal(target))
-        return NULL;
-    UA_NodeId id = UA_NodePointer_toNodeId(target);
-    return UA_NODESTORE_GET(server, &id);
+const UA_ReferenceTarget *
+UA_NodeReferenceKind_findTarget(const UA_NodeReferenceKind *rk,
+                                const UA_ExpandedNodeId *targetId) {
+    UA_NodePointer targetP = UA_NodePointer_fromExpandedNodeId(targetId);
+    if(rk->hasRefTree) {
+        /* Return from the tree */
+        UA_ReferenceTargetTreeElem tmpTarget;
+        tmpTarget.target.targetId = targetP;
+        tmpTarget.targetIdHash = UA_ExpandedNodeId_hash(targetId);
+        UA_ReferenceTargetTreeElem *result =
+            ZIP_FIND(UA_ReferenceIdTree, (UA_ReferenceIdTree*)
+                     (uintptr_t)&rk->targets.tree.idRoot, &tmpTarget);
+        if(result)
+            return &result->target;
+    } else {
+        /* Return from the array */
+        for(size_t i = 0; i < rk->targetsSize; i++) {
+            if(UA_NodePointer_equal(targetP, rk->targets.array[i].targetId))
+                return &rk->targets.array[i];
+        }
+    }
+    return NULL;
 }
 
 /* General node handling methods. There is no UA_Node_new() method here.
@@ -377,8 +394,20 @@ void UA_Node_clear(UA_Node *node) {
     UA_NodeHead *head = &node->head;
     UA_NodeId_clear(&head->nodeId);
     UA_QualifiedName_clear(&head->browseName);
-    UA_LocalizedText_clear(&head->displayName);
-    UA_LocalizedText_clear(&head->description);
+
+    UA_LocalizedTextListEntry *lt;
+
+    while((lt = head->displayName)) {
+        head->displayName = lt->next;
+        UA_LocalizedText_clear(&lt->localizedText);
+        UA_free(lt);
+    }
+
+    while((lt = head->description)) {
+        head->description = lt->next;
+        UA_LocalizedText_clear(&lt->localizedText);
+        UA_free(lt);
+    }
 
     /* Delete unique content of the nodeclass */
     switch(head->nodeClass) {
@@ -430,13 +459,15 @@ UA_CommonVariableNode_copy(const UA_VariableNode *src, UA_VariableNode *dst) {
     dst->arrayDimensionsSize = src->arrayDimensionsSize;
     retval = UA_NodeId_copy(&src->dataType, &dst->dataType);
     dst->valueRank = src->valueRank;
+    dst->valueBackend = src->valueBackend;
     dst->valueSource = src->valueSource;
     if(src->valueSource == UA_VALUESOURCE_DATA) {
         retval |= UA_DataValue_copy(&src->value.data.value,
                                     &dst->value.data.value);
         dst->value.data.callback = src->value.data.callback;
-    } else
+    } else {
         dst->value.dataSource = src->value.dataSource;
+    }
     return retval;
 }
 
@@ -496,6 +527,15 @@ UA_ViewNode_copy(const UA_ViewNode *src, UA_ViewNode *dst) {
     return UA_STATUSCODE_GOOD;
 }
 
+static void *
+copyTarget(void *context, UA_ReferenceTargetTreeElem *elm) {
+    UA_NodeReferenceKind *drefs = (UA_NodeReferenceKind*)context;
+    return (void*)(uintptr_t)
+        addReferenceTargetToTree(drefs, elm->target.targetId,
+                                 elm->targetIdHash,
+                                 elm->target.targetNameHash);
+}
+
 UA_StatusCode
 UA_Node_copy(const UA_Node *src, UA_Node *dst) {
     const UA_NodeHead *srchead = &src->head;
@@ -506,8 +546,37 @@ UA_Node_copy(const UA_Node *src, UA_Node *dst) {
     /* Copy standard content */
     UA_StatusCode retval = UA_NodeId_copy(&srchead->nodeId, &dsthead->nodeId);
     retval |= UA_QualifiedName_copy(&srchead->browseName, &dsthead->browseName);
-    retval |= UA_LocalizedText_copy(&srchead->displayName, &dsthead->displayName);
-    retval |= UA_LocalizedText_copy(&srchead->description, &dsthead->description);
+
+    /* Copy the display name in several languages */
+    for(UA_LocalizedTextListEntry *lt = srchead->displayName; lt != NULL; lt = lt->next) {
+        UA_LocalizedTextListEntry *newEntry = (UA_LocalizedTextListEntry *)
+            UA_calloc(1, sizeof(UA_LocalizedTextListEntry));
+        if(!newEntry) {
+            retval |= UA_STATUSCODE_BADOUTOFMEMORY;
+            break;
+        }
+        retval |= UA_LocalizedText_copy(&lt->localizedText, &newEntry->localizedText);
+
+        /* Add to the linked list possibly in reverse order */
+        newEntry->next = dsthead->displayName;
+        dsthead->displayName = newEntry;
+    }
+
+    /* Copy the description in several languages */
+    for(UA_LocalizedTextListEntry *lt = srchead->description; lt != NULL; lt = lt->next) {
+        UA_LocalizedTextListEntry *newEntry = (UA_LocalizedTextListEntry *)
+            UA_calloc(1, sizeof(UA_LocalizedTextListEntry));
+        if(!newEntry) {
+            retval |= UA_STATUSCODE_BADOUTOFMEMORY;
+            break;
+        }
+        retval |= UA_LocalizedText_copy(&lt->localizedText, &newEntry->localizedText);
+
+        /* Add to the linked list possibly in reverse order */
+        newEntry->next = dsthead->description;
+        dsthead->description= newEntry;
+    }
+
     dsthead->writeMask = srchead->writeMask;
     dsthead->context = srchead->context;
     dsthead->constructed = srchead->constructed;
@@ -538,14 +607,36 @@ UA_Node_copy(const UA_Node *src, UA_Node *dst) {
             drefs->hasRefTree = srefs->hasRefTree; /* initially empty */
 
             /* Copy all the targets */
-            const UA_ReferenceTarget *t = NULL;
-            while((t = UA_NodeReferenceKind_iterate(srefs, t))) {
-                retval = addReferenceTarget(drefs, t->targetId, t->targetNameHash);
-                if(retval != UA_STATUSCODE_GOOD) {
+            if(!srefs->hasRefTree) {
+                drefs->targets.array = (UA_ReferenceTarget*)
+                    UA_malloc(sizeof(UA_ReferenceTarget) * srefs->targetsSize);
+                if(!drefs->targets.array) {
                     UA_Node_clear(dst);
-                    return retval;
+                    return UA_STATUSCODE_BADOUTOFMEMORY;
+                }
+                for(size_t j = 0; j < srefs->targetsSize; j++) {
+                    drefs->targets.array[j].targetNameHash =
+                        srefs->targets.array[j].targetNameHash;
+                    retval = UA_NodePointer_copy(srefs->targets.array[j].targetId,
+                                                 &drefs->targets.array[j].targetId);
+                    drefs->targetsSize++; /* avoid that targetsSize == 0 in error case */
+                    if(retval != UA_STATUSCODE_GOOD) {
+                        UA_Node_clear(dst);
+                        return retval;
+                    }
+                }
+            } else {
+                void *res = ZIP_ITER(UA_ReferenceIdTree,
+                                     (UA_ReferenceIdTree*)&srefs->targets.tree.idRoot,
+                                     copyTarget, drefs);
+                if(res != NULL) {
+                    UA_Node_clear(dst);
+                    return (UA_StatusCode)(uintptr_t)res;
                 }
             }
+
+            UA_assert(srefs->targetsSize > 0);
+            UA_assert(srefs->targetsSize == drefs->targetsSize);
         }
     }
 
@@ -640,14 +731,17 @@ copyStandardAttributes(UA_NodeHead *head, const UA_NodeAttributes *attr) {
     /* UA_QualifiedName_copy(&item->browseName, &node->browseName); */
 
     head->writeMask = attr->writeMask;
-    UA_StatusCode retval = UA_LocalizedText_copy(&attr->description, &head->description);
+    UA_StatusCode retval = UA_Node_insertOrUpdateDescription(head, &attr->description);
     /* The new nodeset format has optional display names:
      * https://github.com/open62541/open62541/issues/2627. If the display name
      * is NULL, take the name part of the browse name */
-    if(attr->displayName.text.length == 0)
-        retval |= UA_String_copy(&head->browseName.name, &head->displayName.text);
-    else
-        retval |= UA_LocalizedText_copy(&attr->displayName, &head->displayName);
+    if(attr->displayName.text.length == 0) {
+        UA_LocalizedText lt;
+        UA_LocalizedText_init(&lt);
+        lt.text = head->browseName.name;
+        retval |= UA_Node_insertOrUpdateDisplayName(head, &lt);
+    } else
+        retval |= UA_Node_insertOrUpdateDisplayName(head, &attr->displayName);
     return retval;
 }
 
@@ -801,66 +895,62 @@ UA_Node_setAttributes(UA_Node *node, const void *attributes, const UA_DataType *
 /*********************/
 
 static UA_StatusCode
-addReferenceTarget(UA_NodeReferenceKind *rk, UA_NodePointer targetId,
-                   UA_UInt32 targetNameHash) {
-    /* Insert into array */
-    if(!rk->hasRefTree) {
-        UA_ReferenceTarget *newRefs = (UA_ReferenceTarget*)
-            UA_realloc(rk->targets.array,
-                       sizeof(UA_ReferenceTarget) * (rk->targetsSize + 1));
-        if(!newRefs)
-            return UA_STATUSCODE_BADOUTOFMEMORY;
-        rk->targets.array = newRefs;
-
-        UA_StatusCode retval =
-            UA_NodePointer_copy(targetId,
-                                &rk->targets.array[rk->targetsSize].targetId);
-        rk->targets.array[rk->targetsSize].targetNameHash = targetNameHash;
-        if(retval != UA_STATUSCODE_GOOD) {
-            if(rk->targetsSize == 0) {
-                UA_free(rk->targets.array);
-                rk->targets.array = NULL;
-            }
-            return retval;
-        }
-        rk->targetsSize++;
-        return UA_STATUSCODE_GOOD;
-    }
-
-    /* Insert into tree */
+addReferenceTargetToTree(UA_NodeReferenceKind *rk, UA_NodePointer targetId,
+                         UA_UInt32 targetIdHash, UA_UInt32 targetNameHash) {
     UA_ReferenceTargetTreeElem *entry = (UA_ReferenceTargetTreeElem*)
         UA_malloc(sizeof(UA_ReferenceTargetTreeElem));
     if(!entry)
         return UA_STATUSCODE_BADOUTOFMEMORY;
 
-    UA_StatusCode retval =
-        UA_NodePointer_copy(targetId, &entry->target.targetId);
-    if(retval != UA_STATUSCODE_GOOD) {
+    UA_StatusCode res = UA_NodePointer_copy(targetId, &entry->target.targetId);
+    if(res != UA_STATUSCODE_GOOD) {
         UA_free(entry);
+        return res;
+    }
+
+    entry->targetIdHash = targetIdHash;
+    entry->target.targetNameHash = targetNameHash;
+
+    ZIP_INSERT(UA_ReferenceIdTree,
+               (UA_ReferenceIdTree*)&rk->targets.tree.idRoot, entry);
+    ZIP_INSERT(UA_ReferenceNameTree,
+               (UA_ReferenceNameTree*)&rk->targets.tree.nameRoot, entry);
+
+    rk->targetsSize++;
+    return UA_STATUSCODE_GOOD;
+}
+
+static UA_StatusCode
+addReferenceTarget(UA_NodeReferenceKind *rk, UA_NodePointer targetId,
+                   UA_UInt32 targetNameHash) {
+    /* Insert into tree */
+    if(rk->hasRefTree) {
+        UA_ExpandedNodeId en = UA_NodePointer_toExpandedNodeId(targetId);
+        return addReferenceTargetToTree(rk, targetId, UA_ExpandedNodeId_hash(&en),
+                                        targetNameHash);
+    }
+
+    /* Insert to the array */
+    UA_ReferenceTarget *newRefs = (UA_ReferenceTarget*)
+        UA_realloc(rk->targets.array,
+                   sizeof(UA_ReferenceTarget) * (rk->targetsSize + 1));
+    if(!newRefs)
+        return UA_STATUSCODE_BADOUTOFMEMORY;
+    rk->targets.array = newRefs;
+
+    UA_StatusCode retval =
+        UA_NodePointer_copy(targetId,
+                            &rk->targets.array[rk->targetsSize].targetId);
+    rk->targets.array[rk->targetsSize].targetNameHash = targetNameHash;
+    if(retval != UA_STATUSCODE_GOOD) {
+        if(rk->targetsSize == 0) {
+            UA_free(rk->targets.array);
+            rk->targets.array = NULL;
+        }
         return retval;
     }
 
-    /* <-- The point of no return --> */
-
-    UA_ExpandedNodeId en = UA_NodePointer_toExpandedNodeId(targetId);
-    entry->targetIdHash = UA_ExpandedNodeId_hash(&en);
-    entry->target.targetNameHash = targetNameHash;
-
-    /* Insert to the id lookup binary search tree. Only the root is kept in refs
-     * to save space. */
-    struct aa_head _refIdTree = refIdTree;
-    _refIdTree.root = rk->targets.tree.idTreeRoot;
-    aa_insert(&_refIdTree, entry);
-    rk->targets.tree.idTreeRoot = _refIdTree.root;
-
-    /* Insert to the name lookup binary search tree */
-    struct aa_head _refNameTree = refNameTree;
-    _refNameTree.root = rk->targets.tree.nameTreeRoot;
-    aa_insert(&_refNameTree, entry);
-    rk->targets.tree.nameTreeRoot = _refNameTree.root;
-
     rk->targetsSize++;
-
     return UA_STATUSCODE_GOOD;
 }
 
@@ -878,14 +968,13 @@ addReferenceKind(UA_NodeHead *head, UA_Byte refTypeIndex, UA_Boolean isForward,
     memset(newRef, 0, sizeof(UA_NodeReferenceKind));
     newRef->referenceTypeIndex = refTypeIndex;
     newRef->isInverse = !isForward;
-    UA_StatusCode retval =
-        addReferenceTarget(newRef, target, targetBrowseNameHash);
-    if(retval != UA_STATUSCODE_GOOD) {
+    UA_StatusCode res = addReferenceTarget(newRef, target, targetBrowseNameHash);
+    if(res != UA_STATUSCODE_GOOD) {
         if(head->referencesSize == 0) {
             UA_free(head->references);
             head->references = NULL;
         }
-        return retval;
+        return res;
     }
 
     head->referencesSize++;
@@ -929,9 +1018,6 @@ UA_Node_addReference(UA_Node *node, UA_Byte refTypeIndex, UA_Boolean isForward,
 UA_StatusCode
 UA_Node_deleteReference(UA_Node *node, UA_Byte refTypeIndex, UA_Boolean isForward,
                         const UA_ExpandedNodeId *targetNodeId) {
-    struct aa_head _refIdTree = refIdTree;
-    struct aa_head _refNameTree = refNameTree;
-
     UA_NodeHead *head = &node->head;
     for(size_t i = 0; i < head->referencesSize; i++) {
         UA_NodeReferenceKind *refs = &head->references[i];
@@ -968,19 +1054,15 @@ UA_Node_deleteReference(UA_Node *node, UA_Byte refTypeIndex, UA_Boolean isForwar
             /* Remove the last target. Remove the ReferenceKind below */
             UA_free(refs->targets.array);
         } else {
-            /* Remove from the tree */
-            _refIdTree.root = refs->targets.tree.idTreeRoot;
-            aa_remove(&_refIdTree, target);
-            refs->targets.tree.idTreeRoot = _refIdTree.root;
-
-            _refNameTree.root = refs->targets.tree.nameTreeRoot;
-            aa_remove(&_refNameTree, target);
-            refs->targets.tree.nameTreeRoot = _refNameTree.root;
-
+            UA_ReferenceTargetTreeElem *elem = (UA_ReferenceTargetTreeElem*)target;
+            ZIP_REMOVE(UA_ReferenceIdTree,
+                       (UA_ReferenceIdTree*)&refs->targets.tree.idRoot, elem);
+            ZIP_REMOVE(UA_ReferenceNameTree,
+                       (UA_ReferenceNameTree*)&refs->targets.tree.nameRoot, elem);
             UA_NodePointer_clear(&target->targetId);
             UA_free(target);
-            if(refs->targets.tree.idTreeRoot)
-                return UA_STATUSCODE_GOOD; /* At least one target remains */
+            if(refs->targetsSize > 0)
+                return UA_STATUSCODE_GOOD;
         }
 
         /* No targets remaining. Remove the ReferenceKind. */
@@ -1009,7 +1091,6 @@ UA_Node_deleteReference(UA_Node *node, UA_Byte refTypeIndex, UA_Boolean isForwar
 void
 UA_Node_deleteReferencesSubset(UA_Node *node, const UA_ReferenceTypeSet *keepSet) {
     UA_NodeHead *head = &node->head;
-    struct aa_head _refIdTree = refIdTree;
     for(size_t i = 0; i < head->referencesSize; i++) {
         /* Keep the references of this type? */
         UA_NodeReferenceKind *refs = &head->references[i];
@@ -1023,15 +1104,9 @@ UA_Node_deleteReferencesSubset(UA_Node *node, const UA_ReferenceTypeSet *keepSet
                 UA_NodePointer_clear(&refs->targets.array[j].targetId);
             UA_free(refs->targets.array);
         } else {
-            _refIdTree.root = refs->targets.tree.idTreeRoot;
-            while(_refIdTree.root) {
-                UA_ReferenceTargetTreeElem *elem = (UA_ReferenceTargetTreeElem*)
-                    ((uintptr_t)_refIdTree.root -
-                     offsetof(UA_ReferenceTargetTreeElem, idTreeEntry));
-                aa_remove(&_refIdTree, elem);
-                UA_NodePointer_clear(&elem->target.targetId);
-                UA_free(elem);
-            }
+            ZIP_ITER(UA_ReferenceIdTree,
+                     (UA_ReferenceIdTree*)&refs->targets.tree.idRoot,
+                     removeTreeEntry, NULL);
         }
 
         /* Move last references-kind entry to this position. Don't memcpy over
@@ -1062,3 +1137,69 @@ void UA_Node_deleteReferences(UA_Node *node) {
     UA_ReferenceTypeSet_init(&noRefs);
     UA_Node_deleteReferencesSubset(node, &noRefs);
 }
+
+static UA_StatusCode
+UA_Node_insertOrUpdateLocale(UA_LocalizedTextListEntry **root,
+                             const UA_LocalizedText *value) {
+    UA_StatusCode res;
+    UA_LocalizedTextListEntry *lt, *prev = NULL;
+    for(lt = *root; lt != NULL; prev = lt, lt = lt->next) {
+        if(!UA_String_equal(&value->locale, &lt->localizedText.locale))
+            continue;
+
+        /* No text -> remove the entry for this locale */
+        if(value->text.length == 0) {
+            if(prev == NULL)
+                *root = lt->next;
+            else
+                prev->next = lt->next;
+            UA_LocalizedText_clear(&lt->localizedText);
+            UA_free(lt);
+            return UA_STATUSCODE_GOOD;
+        }
+
+        /* First make a copy of the text, if this succeeds replace the old
+         * version */
+        UA_String tmp;
+        res = UA_String_copy(&value->text, &tmp);
+        if(res != UA_STATUSCODE_GOOD)
+            return res;
+
+        UA_String_clear(&lt->localizedText.text);
+        lt->localizedText.text = tmp;
+        return UA_STATUSCODE_GOOD;
+    }
+
+    /* The locale does not exist so far */
+
+    /* Do nothing if a non-existing locale should be removed */
+    if(value->text.length == 0)
+        return UA_STATUSCODE_GOOD;
+
+    /* Add a new localized text */
+    lt = (UA_LocalizedTextListEntry *)UA_malloc(sizeof(UA_LocalizedTextListEntry));
+    if(!lt)
+        return UA_STATUSCODE_BADOUTOFMEMORY;
+
+    res = UA_LocalizedText_copy(value, &lt->localizedText);
+    if(res != UA_STATUSCODE_GOOD) {
+        UA_free(lt);
+        return res;
+    }
+
+    lt->next = *root;
+    *root = lt;
+    return UA_STATUSCODE_GOOD;
+}
+
+UA_StatusCode
+UA_Node_insertOrUpdateDisplayName(UA_NodeHead *head,
+                                  const UA_LocalizedText *value) {
+    return UA_Node_insertOrUpdateLocale(&head->displayName, value);
+}
+
+UA_StatusCode
+UA_Node_insertOrUpdateDescription(UA_NodeHead *head,
+                                  const UA_LocalizedText *value) {
+    return UA_Node_insertOrUpdateLocale(&head->description, value);
+}
