@@ -8,6 +8,7 @@
 #include "ua_server_internal.h"
 #include "ua_session.h"
 #include "ua_subscription.h"
+#include "itoa.h"
 
 #ifdef UA_ENABLE_DIAGNOSTICS
 
@@ -33,7 +34,7 @@ fillSubscriptionDiagnostics(UA_Subscription *sub,
     diag->maxKeepAliveCount = sub->maxKeepAliveCount;
     diag->maxLifetimeCount = sub->lifeTimeCount;
     diag->maxNotificationsPerPublish = sub->notificationsPerPublish;
-    diag->publishingEnabled = sub->publishingEnabled;
+    diag->publishingEnabled = (sub->state > UA_SUBSCRIPTIONSTATE_ENABLED_NOPUBLISH);
     diag->modifyCount = sub->modifyCount;
     diag->enableCount = sub->enableCount;
     diag->disableCount = sub->disableCount;
@@ -73,22 +74,16 @@ readSubscriptionDiagnostics(UA_Server *server,
                             const UA_NodeId *nodeId, void *nodeContext,
                             UA_Boolean sourceTimestamp,
                             const UA_NumericRange *range, UA_DataValue *value) {
-    UA_LOCK(&server->serviceMutex);
-
     /* Check the Subscription pointer */
     UA_Subscription *sub = (UA_Subscription*)nodeContext;
-    if(!sub) {
-        UA_UNLOCK(&server->serviceMutex);
+    if(!sub)
         return UA_STATUSCODE_BADINTERNALERROR;
-    }
 
     /* Read the BrowseName */
     UA_QualifiedName bn;
-    UA_StatusCode res = readWithReadValue(server, nodeId, UA_ATTRIBUTEID_BROWSENAME, &bn);
-    if(res != UA_STATUSCODE_GOOD) {
-        UA_UNLOCK(&server->serviceMutex);
+    UA_StatusCode res = UA_Server_readBrowseName(server, *nodeId, &bn);
+    if(res != UA_STATUSCODE_GOOD)
         return res;
-    }
 
     /* Set the value */
     UA_SubscriptionDiagnosticsDataType sddt;
@@ -118,66 +113,10 @@ readSubscriptionDiagnostics(UA_Server *server,
 
     UA_SubscriptionDiagnosticsDataType_clear(&sddt);
     UA_QualifiedName_clear(&bn);
-
-    UA_UNLOCK(&server->serviceMutex);
     return res;
 }
 
-/* If the nodeContext == NULL, return all subscriptions in the server.
- * Otherwise only for the current session. */
-static UA_StatusCode
-readSubscriptionDiagnosticsArrayLocked(UA_Server *server,
-                                       const UA_NodeId *sessionId, void *sessionContext,
-                                       const UA_NodeId *nodeId, void *nodeContext,
-                                       UA_Boolean sourceTimestamp,
-                                       const UA_NumericRange *range, UA_DataValue *value) {
-    UA_LOCK_ASSERT(&server->serviceMutex, 1);
-
-    /* Get the current session */
-    size_t sdSize = 0;
-    UA_Session *session = NULL;
-    session_list_entry *sentry;
-    if(nodeContext) {
-        session = UA_Server_getSessionById(server, sessionId);
-        if(!session)
-            return UA_STATUSCODE_BADINTERNALERROR;
-        sdSize = session->subscriptionsSize;
-    } else {
-        LIST_FOREACH(sentry, &server->sessions, pointers) {
-            sdSize += sentry->session.subscriptionsSize;
-        }
-    }
-
-    /* Allocate the output array */
-    UA_SubscriptionDiagnosticsDataType *sd = (UA_SubscriptionDiagnosticsDataType*)
-        UA_Array_new(sdSize, &UA_TYPES[UA_TYPES_SUBSCRIPTIONDIAGNOSTICSDATATYPE]);
-    if(!sd)
-        return UA_STATUSCODE_BADOUTOFMEMORY;
-
-    /* Collect the statistics */
-    size_t i = 0;
-    UA_Subscription *sub;
-    if(session) {
-        TAILQ_FOREACH(sub, &session->subscriptions, sessionListEntry) {
-            fillSubscriptionDiagnostics(sub, &sd[i]);
-            i++;
-        }
-    } else {
-        LIST_FOREACH(sentry, &server->sessions, pointers) {
-            TAILQ_FOREACH(sub, &sentry->session.subscriptions, sessionListEntry) {
-                fillSubscriptionDiagnostics(sub, &sd[i]);
-                i++;
-            }
-        }
-    }
-
-    /* Set the output */
-    value->hasValue = true;
-    UA_Variant_setArray(&value->value, sd, sdSize,
-                        &UA_TYPES[UA_TYPES_SUBSCRIPTIONDIAGNOSTICSDATATYPE]);
-    return UA_STATUSCODE_GOOD;
-}
-
+/* Return all subscriptions in the server. */
 UA_StatusCode
 readSubscriptionDiagnosticsArray(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
@@ -185,11 +124,39 @@ readSubscriptionDiagnosticsArray(UA_Server *server,
                                  UA_Boolean sourceTimestamp,
                                  const UA_NumericRange *range, UA_DataValue *value) {
     UA_LOCK(&server->serviceMutex);
-    UA_StatusCode res = readSubscriptionDiagnosticsArrayLocked(
-        server, sessionId, sessionContext, nodeId, nodeContext,
-        sourceTimestamp, range, value);
+
+    /* Get the current session */
+    size_t sdSize = 0;
+    session_list_entry *sentry;
+    LIST_FOREACH(sentry, &server->sessions, pointers) {
+        sdSize += sentry->session.subscriptionsSize;
+    }
+
+    /* Allocate the output array */
+    UA_SubscriptionDiagnosticsDataType *sd = (UA_SubscriptionDiagnosticsDataType*)
+        UA_Array_new(sdSize, &UA_TYPES[UA_TYPES_SUBSCRIPTIONDIAGNOSTICSDATATYPE]);
+    if(!sd) {
+        UA_UNLOCK(&server->serviceMutex);
+        return UA_STATUSCODE_BADOUTOFMEMORY;
+    }
+
+    /* Collect the statistics */
+    size_t i = 0;
+    UA_Subscription *sub;
+    LIST_FOREACH(sentry, &server->sessions, pointers) {
+        TAILQ_FOREACH(sub, &sentry->session.subscriptions, sessionListEntry) {
+            fillSubscriptionDiagnostics(sub, &sd[i]);
+            i++;
+        }
+    }
+
+    /* Set the output */
+    value->hasValue = true;
+    UA_Variant_setArray(&value->value, sd, sdSize,
+                        &UA_TYPES[UA_TYPES_SUBSCRIPTIONDIAGNOSTICSDATATYPE]);
+
     UA_UNLOCK(&server->serviceMutex);
-    return res;
+    return UA_STATUSCODE_GOOD;
 }
 
 void
@@ -201,7 +168,7 @@ createSubscriptionObject(UA_Server *server, UA_Session *session,
     UA_NodeId hasComponent = UA_NODEID_NUMERIC(0, UA_NS0ID_HASCOMPONENT);
 
     char subIdStr[32];
-    snprintf(subIdStr, 32, "%u", sub->subscriptionId);
+    itoaUnsigned(sub->subscriptionId, subIdStr, 10);
 
     /* Find the NodeId of the SubscriptionDiagnosticsArray */
     UA_BrowsePath bp;
@@ -224,27 +191,28 @@ createSubscriptionObject(UA_Server *server, UA_Session *session,
     UA_NodeId refId = UA_NODEID_NUMERIC(0, UA_NS0ID_HASCOMPONENT);
     UA_QualifiedName browseName = UA_QUALIFIEDNAME(0, subIdStr);
     UA_NodeId typeId = UA_NODEID_NUMERIC(0, UA_NS0ID_SUBSCRIPTIONDIAGNOSTICSTYPE);
-    UA_NodeId objId = UA_NODEID_NUMERIC(1, 0); /* 0 => assign a random free id */
-    UA_StatusCode res = addNode(server, UA_NODECLASS_VARIABLE,
-                                &objId,
-                                &bpr.targets[0].targetId.nodeId /* parent */,
-                                &refId, browseName, &typeId,
-                                (UA_NodeAttributes*)&var_attr,
-                                &UA_TYPES[UA_TYPES_VARIABLEATTRIBUTES], NULL, &objId);
+    /* Assign a random free NodeId */
+    UA_StatusCode res = addNode(server, UA_NODECLASS_VARIABLE, UA_NODEID_NUMERIC(1, 0),
+                                bpr.targets[0].targetId.nodeId,
+                                refId, browseName, typeId, &var_attr,
+                                &UA_TYPES[UA_TYPES_VARIABLEATTRIBUTES], NULL,
+                                &sub->ns0Id);
     UA_CHECK_STATUS(res, goto cleanup);
 
     /* Add a second reference from the overall SubscriptionDiagnosticsArray variable */
     const UA_NodeId subDiagArray =
         UA_NODEID_NUMERIC(0, UA_NS0ID_SERVER_SERVERDIAGNOSTICS_SUBSCRIPTIONDIAGNOSTICSARRAY);
-    res = addRef(server, session,  &subDiagArray, &refId, &objId, true);
+    res = addRefWithSession(server, session,  &subDiagArray, &refId, &sub->ns0Id, true);
+    if(res != UA_STATUSCODE_GOOD)
+        goto cleanup;
 
     /* Get all children (including the variable itself) and set the contenxt + callback */
     res = referenceTypeIndices(server, &hasComponent, &refTypes, false);
-    if(res != UA_STATUSCODE_GOOD)
-        goto cleanup;
-    res = browseRecursive(server, 1, &objId,
-                          UA_BROWSEDIRECTION_FORWARD, &refTypes,
-                          UA_NODECLASS_VARIABLE, true, &childrenSize, &children);
+    if(UA_LIKELY(res == UA_STATUSCODE_GOOD)) {
+        res = browseRecursive(server, 1, &sub->ns0Id,
+                              UA_BROWSEDIRECTION_FORWARD, &refTypes,
+                              UA_NODECLASS_VARIABLE, true, &childrenSize, &children);
+    }
     if(res != UA_STATUSCODE_GOOD)
         goto cleanup;
 
@@ -260,18 +228,49 @@ createSubscriptionObject(UA_Server *server, UA_Session *session,
  cleanup:
     UA_BrowsePathResult_clear(&bpr);
     if(res != UA_STATUSCODE_GOOD) {
-        UA_LOG_WARNING_SESSION(&server->config.logger, session,
+        UA_LOG_WARNING_SESSION(server->config.logging, session,
                                "Creating the subscription diagnostics object failed "
                                "with StatusCode %s", UA_StatusCode_name(res));
     }
 }
 
-#endif /* UA_ENABLE_SUBSCRIPTIONS */
-
 /***********************/
 /* Session Diagnostics */
 /***********************/
 
+static UA_StatusCode
+setSessionSubscriptionDiagnostics(UA_Server *server, UA_Session *session,
+                                  UA_DataValue *value) {
+    UA_LOCK_ASSERT(&server->serviceMutex, 1);
+
+    /* Get the current session */
+    size_t sdSize = session->subscriptionsSize;
+
+    /* Allocate the output array */
+    UA_SubscriptionDiagnosticsDataType *sd = (UA_SubscriptionDiagnosticsDataType*)
+        UA_Array_new(sdSize, &UA_TYPES[UA_TYPES_SUBSCRIPTIONDIAGNOSTICSDATATYPE]);
+    if(!sd) {
+        UA_UNLOCK(&server->serviceMutex);
+        return UA_STATUSCODE_BADOUTOFMEMORY;
+    }
+
+    /* Collect the statistics */
+    size_t i = 0;
+    UA_Subscription *sub;
+    TAILQ_FOREACH(sub, &session->subscriptions, sessionListEntry) {
+        fillSubscriptionDiagnostics(sub, &sd[i]);
+        i++;
+    }
+
+    /* Set the output */
+    value->hasValue = true;
+    UA_Variant_setArray(&value->value, sd, sdSize,
+                        &UA_TYPES[UA_TYPES_SUBSCRIPTIONDIAGNOSTICSDATATYPE]);
+    return UA_STATUSCODE_GOOD;
+}
+
+#endif /* UA_ENABLE_SUBSCRIPTIONS */
+
 static void
 setSessionDiagnostics(UA_Session *session, UA_SessionDiagnosticsDataType *sd) {
     UA_SessionDiagnosticsDataType_copy(&session->diagnostics, sd);
@@ -316,6 +315,8 @@ readSessionDiagnosticsArray(UA_Server *server,
     if(!sd)
         return UA_STATUSCODE_BADOUTOFMEMORY;
 
+    UA_LOCK(&server->serviceMutex);
+
     /* Collect the statistics */
     size_t i = 0;
     session_list_entry *session;
@@ -328,6 +329,8 @@ readSessionDiagnosticsArray(UA_Server *server,
     value->hasValue = true;
     UA_Variant_setArray(&value->value, sd, server->sessionCount,
                         &UA_TYPES[UA_TYPES_SESSIONDIAGNOSTICSDATATYPE]);
+
+    UA_UNLOCK(&server->serviceMutex);
     return UA_STATUSCODE_GOOD;
 }
 
@@ -336,6 +339,7 @@ setSessionSecurityDiagnostics(UA_Session *session,
                               UA_SessionSecurityDiagnosticsDataType *sd) {
     UA_SessionSecurityDiagnosticsDataType_copy(&session->securityDiagnostics, sd);
     UA_NodeId_copy(&session->sessionId, &sd->sessionId);
+    UA_String_copy(&session->clientUserIdOfSession, &sd->clientUserIdOfSession);
     UA_SecureChannel *channel = session->header.channel;
     if(channel) {
         UA_ByteString_copy(&channel->remoteCertificate, &sd->clientCertificate);
@@ -355,12 +359,12 @@ readSessionDiagnostics(UA_Server *server,
     UA_LOCK(&server->serviceMutex);
 
     /* Get the Session */
-    UA_Session *session = UA_Server_getSessionById(server, sessionId);
+    UA_Session *session = getSessionById(server, sessionId);
     if(!session) {
         UA_UNLOCK(&server->serviceMutex);
         return UA_STATUSCODE_BADINTERNALERROR;
     }
-    
+
     /* Read the BrowseName */
     UA_QualifiedName bn;
     UA_StatusCode res = readWithReadValue(server, nodeId, UA_ATTRIBUTEID_BROWSENAME, &bn);
@@ -382,49 +386,46 @@ readSessionDiagnostics(UA_Server *server,
     size_t memberOffset;
     UA_Boolean found;
 
+#ifdef UA_ENABLE_SUBSCRIPTIONS
     if(equalBrowseName(&bn.name, "SubscriptionDiagnosticsArray")) {
-        /* Reuse the datasource callback. Forward a non-null nodeContext to
-         * indicate that we want to see only the subscriptions for the current
-         * session. */
-        res = readSubscriptionDiagnosticsArrayLocked(server, sessionId, sessionContext,
-                                                     nodeId, (void*)0x01,
-                                                     sourceTimestamp, range, value);
+        res = setSessionSubscriptionDiagnostics(server, session, value);
         goto cleanup;
-    } else if(equalBrowseName(&bn.name, "SessionDiagnostics")) {
+    }
+#endif
+
+    if(equalBrowseName(&bn.name, "SessionDiagnostics")) {
         setSessionDiagnostics(session, &data.sddt);
         content = &data.sddt;
         type = &UA_TYPES[UA_TYPES_SESSIONDIAGNOSTICSDATATYPE];
-        goto set_value;
     } else if(equalBrowseName(&bn.name, "SessionSecurityDiagnostics")) {
         setSessionSecurityDiagnostics(session, &data.ssddt);
         securityDiagnostics = true;
         content = &data.ssddt;
         type = &UA_TYPES[UA_TYPES_SESSIONSECURITYDIAGNOSTICSDATATYPE];
-        goto set_value;
-    }
-
-    /* Try to find the member in SessionDiagnosticsDataType and
-     * SessionSecurityDiagnosticsDataType */
-    memcpy(memberName, bn.name.data, bn.name.length);
-    memberName[bn.name.length] = 0;
-    found = UA_DataType_getStructMember(&UA_TYPES[UA_TYPES_SESSIONDIAGNOSTICSDATATYPE],
-                                        memberName, &memberOffset, &type, &isArray);
-    if(found) {
-        setSessionDiagnostics(session, &data.sddt);
-        content = (void*)(((uintptr_t)&data.sddt) + memberOffset);
     } else {
-        found = UA_DataType_getStructMember(&UA_TYPES[UA_TYPES_SESSIONSECURITYDIAGNOSTICSDATATYPE],
+        /* Try to find the member in SessionDiagnosticsDataType and
+         * SessionSecurityDiagnosticsDataType */
+        memcpy(memberName, bn.name.data, bn.name.length);
+        memberName[bn.name.length] = 0;
+        found = UA_DataType_getStructMember(&UA_TYPES[UA_TYPES_SESSIONDIAGNOSTICSDATATYPE],
                                             memberName, &memberOffset, &type, &isArray);
-        if(!found) {
-            res = UA_STATUSCODE_BADNOTIMPLEMENTED;
-            goto cleanup;
+        if(found) {
+            setSessionDiagnostics(session, &data.sddt);
+            content = (void*)(((uintptr_t)&data.sddt) + memberOffset);
+        } else {
+            const UA_DataType *dt = &UA_TYPES[UA_TYPES_SESSIONSECURITYDIAGNOSTICSDATATYPE];
+            found = UA_DataType_getStructMember(dt, memberName, &memberOffset,
+                                                &type, &isArray);
+            if(!found) {
+                res = UA_STATUSCODE_BADNOTIMPLEMENTED;
+                goto cleanup;
+            }
+            setSessionSecurityDiagnostics(session, &data.ssddt);
+            securityDiagnostics = true;
+            content = (void*)(((uintptr_t)&data.ssddt) + memberOffset);
         }
-        setSessionSecurityDiagnostics(session, &data.ssddt);
-        securityDiagnostics = true;
-        content = (void*)(((uintptr_t)&data.ssddt) + memberOffset);
     }
 
- set_value:
     if(!isArray) {
         res = UA_Variant_setScalarCopy(&value->value, content, type);
     } else {
@@ -459,6 +460,8 @@ readSessionSecurityDiagnostics(UA_Server *server,
     if(!sd)
         return UA_STATUSCODE_BADOUTOFMEMORY;
 
+    UA_LOCK(&server->serviceMutex);
+
     /* Collect the statistics */
     size_t i = 0;
     session_list_entry *session;
@@ -471,6 +474,8 @@ readSessionSecurityDiagnostics(UA_Server *server,
     value->hasValue = true;
     UA_Variant_setArray(&value->value, sd, server->sessionCount,
                         &UA_TYPES[UA_TYPES_SESSIONSECURITYDIAGNOSTICSDATATYPE]);
+
+    UA_UNLOCK(&server->serviceMutex);
     return UA_STATUSCODE_GOOD;
 }
 
@@ -489,9 +494,8 @@ createSessionObject(UA_Server *server, UA_Session *session) {
     UA_QualifiedName browseName = UA_QUALIFIEDNAME(0, "");
     browseName.name = session->sessionName; /* shallow copy */
     UA_NodeId typeId = UA_NODEID_NUMERIC(0, UA_NS0ID_SESSIONDIAGNOSTICSOBJECTTYPE);
-    UA_StatusCode res = addNode(server, UA_NODECLASS_OBJECT,
-                                &session->sessionId, &parentId, &refId, browseName, &typeId,
-                                (UA_NodeAttributes*)&object_attr,
+    UA_StatusCode res = addNode(server, UA_NODECLASS_OBJECT, session->sessionId,
+                                parentId, refId, browseName, typeId, &object_attr,
                                 &UA_TYPES[UA_TYPES_OBJECTATTRIBUTES], NULL, NULL);
     if(res != UA_STATUSCODE_GOOD)
         goto cleanup;
@@ -514,7 +518,7 @@ createSessionObject(UA_Server *server, UA_Session *session) {
 
  cleanup:
     if(res != UA_STATUSCODE_GOOD) {
-        UA_LOG_WARNING_SESSION(&server->config.logger, session,
+        UA_LOG_WARNING_SESSION(server->config.logging, session,
                                "Creating the session diagnostics object failed "
                                "with StatusCode %s", UA_StatusCode_name(res));
     }
@@ -545,6 +549,8 @@ readDiagnostics(UA_Server *server, const UA_NodeId *sessionId, void *sessionCont
     void *data = NULL;
     const UA_DataType *type = &UA_TYPES[UA_TYPES_UINT32]; /* Default */
 
+    UA_LOCK(&server->serviceMutex);
+
     switch(nodeId->identifier.numeric) {
     case UA_NS0ID_SERVER_SERVERDIAGNOSTICS_SERVERDIAGNOSTICSSUMMARY:
         server->serverDiagnosticsSummary.currentSessionCount =
@@ -589,12 +595,15 @@ readDiagnostics(UA_Server *server, const UA_NodeId *sessionId, void *sessionCont
         data = &server->serverDiagnosticsSummary.rejectedRequestsCount;
         break;
     default:
+        UA_UNLOCK(&server->serviceMutex);
         return UA_STATUSCODE_BADINTERNALERROR;
     }
 
     UA_StatusCode res = UA_Variant_setScalarCopy(&value->value, data, type);
     if(res == UA_STATUSCODE_GOOD)
         value->hasValue = true;
+
+    UA_UNLOCK(&server->serviceMutex);
     return res;
 }
 
