@@ -18,37 +18,56 @@
 
 #ifdef UA_ENABLE_METHODCALLS /* conditional compilation */
 
+#define UA_MAX_METHOD_ARGUMENTS 64
+
+struct GetArgumentsNodeContext {
+    UA_Server *server;
+    UA_String withBrowseName;
+};
+
+static void *
+getArgumentsNodeCallback(void *context, UA_ReferenceTarget *t) {
+    struct GetArgumentsNodeContext *ctx = (struct GetArgumentsNodeContext*)context;
+    const UA_Node *refTarget =
+        UA_NODESTORE_GETFROMREF_SELECTIVE(ctx->server, t->targetId,
+                                          UA_NODEATTRIBUTESMASK_NODECLASS |
+                                          UA_NODEATTRIBUTESMASK_VALUE,
+                                          UA_REFERENCETYPESET_NONE,
+                                          UA_BROWSEDIRECTION_INVALID);
+    if(!refTarget)
+        return NULL;
+    if(refTarget->head.nodeClass == UA_NODECLASS_VARIABLE &&
+       refTarget->head.browseName.namespaceIndex == 0 &&
+       UA_String_equal(&ctx->withBrowseName, &refTarget->head.browseName.name)) {
+        return (void*)(uintptr_t)&refTarget->variableNode;
+    }
+    UA_NODESTORE_RELEASE(ctx->server, refTarget);
+    return NULL;
+}
+
 static const UA_VariableNode *
 getArgumentsVariableNode(UA_Server *server, const UA_NodeHead *head,
                          UA_String withBrowseName) {
     for(size_t i = 0; i < head->referencesSize; ++i) {
-        const UA_NodeReferenceKind *rk = &head->references[i];
-        if(rk->isInverse != false)
+        UA_NodeReferenceKind *rk = &head->references[i];
+        if(rk->isInverse)
             continue;
         if(rk->referenceTypeIndex != UA_REFERENCETYPEINDEX_HASPROPERTY)
             continue;
-        const UA_ReferenceTarget *t = NULL;
-        while((t = UA_NodeReferenceKind_iterate(rk, t))) {
-            const UA_Node *refTarget =
-                UA_NODESTORE_GETFROMREF(server, t->targetId);
-            if(!refTarget)
-                continue;
-            if(refTarget->head.nodeClass == UA_NODECLASS_VARIABLE &&
-               refTarget->head.browseName.namespaceIndex == 0 &&
-               UA_String_equal(&withBrowseName, &refTarget->head.browseName.name)) {
-                return &refTarget->variableNode;
-            }
-            UA_NODESTORE_RELEASE(server, refTarget);
-        }
+        struct GetArgumentsNodeContext ctx;
+        ctx.server = server;
+        ctx.withBrowseName = withBrowseName;
+        return (const UA_VariableNode*)
+            UA_NodeReferenceKind_iterate(rk, getArgumentsNodeCallback, &ctx);
     }
     return NULL;
 }
 
 /* inputArgumentResults has the length request->inputArgumentsSize */
 static UA_StatusCode
-typeCheckArguments(UA_Server *server, UA_Session *session,
-                   const UA_VariableNode *argRequirements, size_t argsSize,
-                   UA_Variant *args, UA_StatusCode *inputArgumentResults) {
+checkAdjustArguments(UA_Server *server, UA_Session *session,
+                     const UA_VariableNode *argRequirements, size_t argsSize,
+                     UA_Variant *args, UA_StatusCode *inputArgumentResults) {
     /* Verify that we have a Variant containing UA_Argument (scalar or array) in
      * the "InputArguments" node */
     if(argRequirements->valueSource != UA_VALUESOURCE_DATA)
@@ -73,15 +92,10 @@ typeCheckArguments(UA_Server *server, UA_Session *session,
     UA_Argument *argReqs = (UA_Argument*)argRequirements->value.data.value.value.data;
     const char *reason;
     for(size_t i = 0; i < argReqsSize; ++i) {
-        if(compatibleValue(server, session, &argReqs[i].dataType, argReqs[i].valueRank,
-                           argReqs[i].arrayDimensionsSize, argReqs[i].arrayDimensions,
-                           &args[i], NULL, &reason))
-            continue;
-
         /* Incompatible value. Try to correct the type if possible. */
         adjustValueType(server, &args[i], &argReqs[i].dataType);
 
-        /* Recheck */
+        /* Check */
         if(!compatibleValue(server, session, &argReqs[i].dataType, argReqs[i].valueRank,
                             argReqs[i].arrayDimensionsSize, argReqs[i].arrayDimensions,
                             &args[i], NULL, &reason)) {
@@ -92,30 +106,6 @@ typeCheckArguments(UA_Server *server, UA_Session *session,
     return retval;
 }
 
-/* inputArgumentResults has the length request->inputArgumentsSize */
-static UA_StatusCode
-validMethodArguments(UA_Server *server, UA_Session *session, const UA_MethodNode *method,
-                     const UA_CallMethodRequest *request,
-                     UA_StatusCode *inputArgumentResults) {
-    /* Get the input arguments node */
-    const UA_VariableNode *inputArguments =
-        getArgumentsVariableNode(server, &method->head, UA_STRING("InputArguments"));
-    if(!inputArguments) {
-        if(request->inputArgumentsSize > 0)
-            return UA_STATUSCODE_BADTOOMANYARGUMENTS;
-        return UA_STATUSCODE_GOOD;
-    }
-
-    /* Verify the request */
-    UA_StatusCode retval =
-        typeCheckArguments(server, session, inputArguments, request->inputArgumentsSize,
-                           request->inputArguments, inputArgumentResults);
-
-    /* Release the input arguments node */
-    UA_NODESTORE_RELEASE(server, (const UA_Node*)inputArguments);
-    return retval;
-}
-
 static const UA_NodeId hasComponentNodeId = {0, UA_NODEIDTYPE_NUMERIC, {UA_NS0ID_HASCOMPONENT}};
 static const UA_NodeId organizedByNodeId = {0, UA_NODEIDTYPE_NUMERIC, {UA_NS0ID_ORGANIZES}};
 static const UA_String namespaceDiModel = UA_STRING_STATIC("http://opcfoundation.org/UA/DI/");
@@ -123,10 +113,96 @@ static const UA_NodeId hasTypeDefinitionNodeId = {0, UA_NODEIDTYPE_NUMERIC, {UA_
 // ns=0 will be replace dynamically. DI-Spec. 1.01: <UAObjectType NodeId="ns=1;i=1005" BrowseName="1:FunctionalGroupType">
 static UA_NodeId functionGroupNodeId = {0, UA_NODEIDTYPE_NUMERIC, {1005}};
 
+static UA_Boolean
+checkMethodReference(const UA_NodeHead *h, UA_ReferenceTypeSet refs,
+                     const UA_ExpandedNodeId *methodId) {
+    for(size_t i = 0; i < h->referencesSize; i++) {
+        const UA_NodeReferenceKind *rk = &h->references[i];
+        if(rk->isInverse)
+            continue;
+        if(!UA_ReferenceTypeSet_contains(&refs, rk->referenceTypeIndex))
+            continue;
+        if(UA_NodeReferenceKind_findTarget(rk, methodId))
+            return true;
+    }
+    return false;
+}
+
+static void *
+iterateFunctionGroupSearch(void *context, UA_ReferenceTarget *t) {
+    UA_Server *server = (UA_Server*)context;
+    if(!UA_NodePointer_isLocal(t->targetId))
+        return NULL;
+
+    UA_NodeId tmpId = UA_NodePointer_toNodeId(t->targetId);
+    if(isNodeInTree_singleRef(server, &tmpId, &functionGroupNodeId,
+                               UA_REFERENCETYPEINDEX_HASSUBTYPE))
+        return (void*)0x01;
+    return NULL;
+}
+
+static UA_StatusCode
+checkFunctionalGroupMethodReference(UA_Server *server, const UA_NodeHead *h,
+                                    const UA_ExpandedNodeId *methodId,
+                                    UA_Boolean *found) {
+    /* Check whether the DI namespace is available */
+    size_t foundNamespace = 0;
+    UA_StatusCode res = getNamespaceByName(server, namespaceDiModel, &foundNamespace);
+    UA_CHECK_STATUS(res, return UA_STATUSCODE_BADMETHODINVALID);
+    functionGroupNodeId.namespaceIndex = (UA_UInt16)foundNamespace;
+
+    UA_ReferenceTypeSet hasTypeDefinitionRefs;
+    res = referenceTypeIndices(server, &hasTypeDefinitionNodeId,
+                               &hasTypeDefinitionRefs, true);
+    UA_CHECK_STATUS(res, return res);
+
+    /* Search for a HasTypeDefinition (or sub-) reference to the FunctionGroupType */
+    UA_Boolean isFunctionGroup = false;
+    for(size_t i = 0; i < h->referencesSize && !isFunctionGroup; ++i) {
+        UA_NodeReferenceKind *rk = &h->references[i];
+        if(rk->isInverse)
+            continue;
+
+        /* Are these HasTypeDefinition references */
+        if(!UA_ReferenceTypeSet_contains(&hasTypeDefinitionRefs, rk->referenceTypeIndex))
+            continue;
+
+        /* Reference points to FunctionGroupType (or sub-type) from the DI
+         * model? */
+        isFunctionGroup =
+            (UA_NodeReferenceKind_iterate(rk, iterateFunctionGroupSearch,
+                                          server) != NULL);
+        if(isFunctionGroup)
+            break;
+    }
+    if(!isFunctionGroup)
+        return UA_STATUSCODE_GOOD;
+
+    /* Search for the called method with reference Organize (or sub-type) from
+     * the parent object */
+    UA_ReferenceTypeSet organizesRefs;
+    res = referenceTypeIndices(server, &organizedByNodeId, &organizesRefs, true);
+    UA_CHECK_STATUS(res, return res);
+    for(size_t k = 0; k < h->referencesSize; ++k) {
+        const UA_NodeReferenceKind *rk = &h->references[k];
+        if(rk->isInverse)
+            continue;
+        if(!UA_ReferenceTypeSet_contains(&organizesRefs, rk->referenceTypeIndex))
+            continue;
+        if(UA_NodeReferenceKind_findTarget(rk, methodId)) {
+            *found = true;
+            break;
+        }
+    }
+    return UA_STATUSCODE_GOOD;
+}
+
 static void
 callWithMethodAndObject(UA_Server *server, UA_Session *session,
                         const UA_CallMethodRequest *request, UA_CallMethodResult *result,
                         const UA_MethodNode *method, const UA_ObjectNode *object) {
+    UA_LOCK_ASSERT(&server->serviceMutex, 1);
+
     /* Verify the object's NodeClass */
     if(object->head.nodeClass != UA_NODECLASS_OBJECT &&
        object->head.nodeClass != UA_NODECLASS_OBJECTTYPE) {
@@ -146,31 +222,26 @@ callWithMethodAndObject(UA_Server *server, UA_Session *session,
         return;
     }
 
-    UA_NodePointer methodP = UA_NodePointer_fromNodeId(&request->methodId);
-
     /* Verify method/object relations. Object must have a hasComponent or a
      * subtype of hasComponent reference to the method node. Therefore, check
      * every reference between the parent object and the method node if there is
      * a hasComponent (or subtype) reference */
+    UA_ExpandedNodeId methodId = UA_EXPANDEDNODEID_NODEID(request->methodId);
     UA_ReferenceTypeSet hasComponentRefs;
-    result->statusCode =
-        referenceTypeIndices(server, &hasComponentNodeId, &hasComponentRefs, true);
-    if(result->statusCode != UA_STATUSCODE_GOOD)
-        return;
+    result->statusCode = referenceTypeIndices(server, &hasComponentNodeId,
+                                              &hasComponentRefs, true);
+    UA_CHECK_STATUS(result->statusCode, return);
+    UA_Boolean found = checkMethodReference(&object->head, hasComponentRefs, &methodId);
 
-    UA_Boolean found = false;
-    for(size_t i = 0; i < object->head.referencesSize && !found; ++i) {
-        const UA_NodeReferenceKind *rk = &object->head.references[i];
-        if(rk->isInverse)
-            continue;
-        if(!UA_ReferenceTypeSet_contains(&hasComponentRefs, rk->referenceTypeIndex))
-            continue;
-        const UA_ReferenceTarget *t = NULL;
-        while((t = UA_NodeReferenceKind_iterate(rk, t))) {
-            if(UA_NodePointer_equal(t->targetId, methodP)) {
-                found = true;
-                break;
-            }
+    if(!found) {
+        /* If the object doesn't have a hasComponent reference to the method node,
+         * check its objectType (and its supertypes). Invoked method can be a component
+         * of objectType and be invoked on this objectType's instance (or on a instance
+         * of one of its subtypes). */
+        const UA_Node *objectType = getNodeType(server, &object->head);
+        if(objectType) {
+            found = checkMethodReference(&objectType->head, hasComponentRefs, &methodId);
+            UA_NODESTORE_RELEASE(server, objectType);
         }
     }
 
@@ -185,69 +256,11 @@ callWithMethodAndObject(UA_Server *server, UA_Session *session,
          * ParameterSet and MethodSet) in (Functional) groups for instance
          * Configuration or Identification. The same Property, Parameter or
          * Method can be referenced from more than one FunctionalGroup. */
-
-        /* Check whether the DI namespace is available */
-        size_t foundNamespace = 0;
-        UA_StatusCode res = getNamespaceByName(server, namespaceDiModel, &foundNamespace);
-        if(res != UA_STATUSCODE_GOOD) {
-            result->statusCode = UA_STATUSCODE_BADMETHODINVALID;
-            return;
-        }
-        functionGroupNodeId.namespaceIndex = (UA_UInt16)foundNamespace;
-
-        UA_ReferenceTypeSet hasTypeDefinitionRefs;
         result->statusCode =
-            referenceTypeIndices(server, &hasTypeDefinitionNodeId,
-                                 &hasTypeDefinitionRefs, true);
-        if(result->statusCode != UA_STATUSCODE_GOOD)
-            return;
-
-        /* Search for a HasTypeDefinition (or sub-) reference in the parent object */
-        for(size_t i = 0; i < object->head.referencesSize && !found; ++i) {
-            const UA_NodeReferenceKind *rk = &object->head.references[i];
-            if(rk->isInverse)
-                continue;
-            if(!UA_ReferenceTypeSet_contains(&hasTypeDefinitionRefs, rk->referenceTypeIndex))
-                continue;
-            
-            /* Verify that the HasTypeDefinition is equal to FunctionGroupType
-             * (or sub-type) from the DI model */
-            const UA_ReferenceTarget *t = NULL;
-            while((t = UA_NodeReferenceKind_iterate(rk, t))) {
-                if(!UA_NodePointer_isLocal(t->targetId))
-                    continue;
-                
-                UA_NodeId tmpId = UA_NodePointer_toNodeId(t->targetId);
-                if(!isNodeInTree_singleRef(server, &tmpId, &functionGroupNodeId,
-                                           UA_REFERENCETYPEINDEX_HASSUBTYPE))
-                    continue;
-
-                /* Search for the called method with reference Organize (or
-                 * sub-type) from the parent object */
-                for(size_t k = 0; k < object->head.referencesSize && !found; ++k) {
-                    const UA_NodeReferenceKind *rkInner = &object->head.references[k];
-                    if(rkInner->isInverse)
-                        continue;
-                    const UA_NodeId * refId = 
-                        UA_NODESTORE_GETREFERENCETYPEID(server, rkInner->referenceTypeIndex);
-                    if(!isNodeInTree_singleRef(server, refId, &organizedByNodeId,
-                                               UA_REFERENCETYPEINDEX_HASSUBTYPE))
-                        continue;
-                    
-                    const UA_ReferenceTarget *t2 = NULL;
-                    while((t2 = UA_NodeReferenceKind_iterate(rkInner, t2))) {
-                        if(UA_NodePointer_equal(t2->targetId, methodP)) {
-                            found = true;
-                            break;
-                        }
-                    }
-                }
-            }
-        }
-        if(!found) {
+            checkFunctionalGroupMethodReference(server, &object->head, &methodId, &found);
+        if(!found && result->statusCode == UA_STATUSCODE_GOOD)
             result->statusCode = UA_STATUSCODE_BADMETHODINVALID;
-            return;
-        }
+        UA_CHECK_STATUS(result->statusCode, return);
     }
 
     /* Verify access rights */
@@ -255,8 +268,9 @@ callWithMethodAndObject(UA_Server *server, UA_Session *session,
     if(session != &server->adminSession) {
         UA_UNLOCK(&server->serviceMutex);
         executable = executable && server->config.accessControl.
-            getUserExecutableOnObject(server, &server->config.accessControl, &session->sessionId,
-                                      session->sessionHandle, &request->methodId, method->head.context,
+            getUserExecutableOnObject(server, &server->config.accessControl,
+                                      &session->sessionId, session->sessionHandle,
+                                      &request->methodId, method->head.context,
                                       &request->objectId, object->head.context);
         UA_LOCK(&server->serviceMutex);
     }
@@ -266,6 +280,19 @@ callWithMethodAndObject(UA_Server *server, UA_Session *session,
         return;
     }
 
+    /* The input arguments are const and not changed. We move the input
+     * arguments to a secondary array that is mutable. This is used for small
+     * adjustments on the type level during the type checking. But it has to be
+     * ensured that the original array can still by _clear'ed after the methods
+     * call. */
+    if(request->inputArgumentsSize > UA_MAX_METHOD_ARGUMENTS) {
+        result->statusCode = UA_STATUSCODE_BADTOOMANYARGUMENTS;
+        return;
+    }
+    UA_Variant mutableInputArgs[UA_MAX_METHOD_ARGUMENTS];
+    memcpy(mutableInputArgs, request->inputArguments,
+           sizeof(UA_Variant) * request->inputArgumentsSize);
+
     /* Allocate the inputArgumentResults array */
     result->inputArgumentResults = (UA_StatusCode*)
         UA_Array_new(request->inputArgumentsSize, &UA_TYPES[UA_TYPES_STATUSCODE]);
@@ -275,9 +302,20 @@ callWithMethodAndObject(UA_Server *server, UA_Session *session,
     }
     result->inputArgumentResultsSize = request->inputArgumentsSize;
 
-    /* Verify Input Arguments */
-    result->statusCode = validMethodArguments(server, session, method, request,
-                                              result->inputArgumentResults);
+    /* Type-check the input arguments */
+    const UA_VariableNode *inputArguments =
+        getArgumentsVariableNode(server, &method->head, UA_STRING("InputArguments"));
+    if(inputArguments) {
+        result->statusCode =
+            checkAdjustArguments(server, session, inputArguments, request->inputArgumentsSize,
+                                 mutableInputArgs, result->inputArgumentResults);
+        UA_NODESTORE_RELEASE(server, (const UA_Node*)inputArguments);
+    } else {
+        if(request->inputArgumentsSize > 0) {
+            result->statusCode = UA_STATUSCODE_BADTOOMANYARGUMENTS;
+            return;
+        }
+    }
 
     /* Return inputArgumentResults only for BADINVALIDARGUMENT */
     if(result->statusCode != UA_STATUSCODE_BADINVALIDARGUMENT) {
@@ -315,7 +353,7 @@ callWithMethodAndObject(UA_Server *server, UA_Session *session,
     result->statusCode = method->method(server, &session->sessionId, session->sessionHandle,
                                         &method->head.nodeId, method->head.context,
                                         &object->head.nodeId, object->head.context,
-                                        request->inputArgumentsSize, request->inputArguments,
+                                        request->inputArgumentsSize, mutableInputArgs,
                                         result->outputArgumentsSize, result->outputArguments);
     UA_LOCK(&server->serviceMutex);
     /* TODO: Verify Output matches the argument definition */
@@ -328,15 +366,29 @@ Operation_CallMethodAsync(UA_Server *server, UA_Session *session, UA_UInt32 requ
                           UA_UInt32 requestHandle, size_t opIndex,
                           UA_CallMethodRequest *opRequest, UA_CallMethodResult *opResult,
                           UA_AsyncResponse **ar) {
-    /* Get the method node */
-    const UA_Node *method = UA_NODESTORE_GET(server, &opRequest->methodId);
+    /* Get the method node. We only need the nodeClass and executable attribute.
+     * Take all forward hasProperty references to get the input/output argument
+     * definition variables. */
+    const UA_Node *method =
+        UA_NODESTORE_GET_SELECTIVE(server, &opRequest->methodId,
+                                   UA_NODEATTRIBUTESMASK_NODECLASS |
+                                   UA_NODEATTRIBUTESMASK_EXECUTABLE,
+                                   UA_REFTYPESET(UA_REFERENCETYPEINDEX_HASPROPERTY),
+                                   UA_BROWSEDIRECTION_FORWARD);
     if(!method) {
-        opResult->statusCode = UA_STATUSCODE_BADNODEIDUNKNOWN;
+        opResult->statusCode = UA_STATUSCODE_BADMETHODINVALID;
         return;
     }
 
-    /* Get the object node */
-    const UA_Node *object = UA_NODESTORE_GET(server, &opRequest->objectId);
+    /* Get the object node. We only need the NodeClass attribute. But take all
+     * references for now.
+     *
+     * TODO: Which references do we need actually? */
+    const UA_Node *object =
+        UA_NODESTORE_GET_SELECTIVE(server, &opRequest->objectId,
+                                   UA_NODEATTRIBUTESMASK_NODECLASS,
+                                   UA_REFERENCETYPESET_ALL,
+                                   UA_BROWSEDIRECTION_BOTH);
     if(!object) {
         opResult->statusCode = UA_STATUSCODE_BADNODEIDUNKNOWN;
         UA_NODESTORE_RELEASE(server, method);
@@ -350,28 +402,6 @@ Operation_CallMethodAsync(UA_Server *server, UA_Session *session, UA_UInt32 requ
         goto cleanup;
     }
 
-    /* Check the NodeClass */
-    if(method->head.nodeClass != UA_NODECLASS_METHOD ||
-       object->head.nodeClass != UA_NODECLASS_OBJECT) {
-        opResult->statusCode = UA_STATUSCODE_BADNODECLASSINVALID;
-        goto cleanup;
-    }
-
-    /* Check the access rights */
-    UA_Boolean executable = method->methodNode.executable;
-    if(session != &server->adminSession) {
-        executable = executable && server->config.accessControl.
-            getUserExecutableOnObject(server, &server->config.accessControl,
-                                      &session->sessionId, session->sessionHandle,
-                                      &opRequest->methodId, method->head.context,
-                                      &opRequest->objectId, object->head.context);
-    }
-
-    if(!executable) {
-        opResult->statusCode = UA_STATUSCODE_BADNOTEXECUTABLE;
-        goto cleanup;
-    }
-
     /* <-- Async method call --> */
 
     /* No AsyncResponse allocated so far */
@@ -399,7 +429,7 @@ void
 Service_CallAsync(UA_Server *server, UA_Session *session, UA_UInt32 requestId,
                   const UA_CallRequest *request, UA_CallResponse *response,
                   UA_Boolean *finished) {
-    UA_LOG_DEBUG_SESSION(&server->config.logger, session, "Processing CallRequestAsync");
+    UA_LOG_DEBUG_SESSION(server->config.logging, session, "Processing CallRequestAsync");
     if(server->config.maxNodesPerMethodCall != 0 &&
         request->methodsToCallSize > server->config.maxNodesPerMethodCall) {
         response->responseHeader.serviceResult = UA_STATUSCODE_BADTOOMANYOPERATIONS;
@@ -433,15 +463,29 @@ Service_CallAsync(UA_Server *server, UA_Session *session, UA_UInt32 requestId,
 static void
 Operation_CallMethod(UA_Server *server, UA_Session *session, void *context,
                      const UA_CallMethodRequest *request, UA_CallMethodResult *result) {
-    /* Get the method node */
-    const UA_Node *method = UA_NODESTORE_GET(server, &request->methodId);
+    /* Get the method node. We only need the nodeClass and executable attribute.
+     * Take all forward hasProperty references to get the input/output argument
+     * definition variables. */
+    const UA_Node *method =
+        UA_NODESTORE_GET_SELECTIVE(server, &request->methodId,
+                                   UA_NODEATTRIBUTESMASK_NODECLASS |
+                                   UA_NODEATTRIBUTESMASK_EXECUTABLE,
+                                   UA_REFTYPESET(UA_REFERENCETYPEINDEX_HASPROPERTY),
+                                   UA_BROWSEDIRECTION_FORWARD);
     if(!method) {
-        result->statusCode = UA_STATUSCODE_BADNODEIDUNKNOWN;
+        result->statusCode = UA_STATUSCODE_BADMETHODINVALID;
         return;
     }
 
-    /* Get the object node */
-    const UA_Node *object = UA_NODESTORE_GET(server, &request->objectId);
+    /* Get the object node. We only need the NodeClass attribute. But take all
+     * references for now.
+     *
+     * TODO: Which references do we need actually? */
+    const UA_Node *object =
+        UA_NODESTORE_GET_SELECTIVE(server, &request->objectId,
+                                   UA_NODEATTRIBUTESMASK_NODECLASS,
+                                   UA_REFERENCETYPESET_ALL,
+                                   UA_BROWSEDIRECTION_BOTH);
     if(!object) {
         result->statusCode = UA_STATUSCODE_BADNODEIDUNKNOWN;
         UA_NODESTORE_RELEASE(server, method);
@@ -459,7 +503,7 @@ Operation_CallMethod(UA_Server *server, UA_Session *session, void *context,
 
 void Service_Call(UA_Server *server, UA_Session *session,
                   const UA_CallRequest *request, UA_CallResponse *response) {
-    UA_LOG_DEBUG_SESSION(&server->config.logger, session, "Processing CallRequest");
+    UA_LOG_DEBUG_SESSION(server->config.logging, session, "Processing CallRequest");
     UA_LOCK_ASSERT(&server->serviceMutex, 1);
 
     if(server->config.maxNodesPerMethodCall != 0 &&
